from __future__ import annotations

import asyncio
import contextlib
import json
import time
from collections.abc import AsyncGenerator
from datetime import timedelta
from typing import TYPE_CHECKING
from unittest.mock import AsyncMock, Mock, patch

import pytest
import pytest_asyncio
from pytest_mock import MockerFixture

from strawberry.subscriptions import GRAPHQL_TRANSPORT_WS_PROTOCOL
from strawberry.subscriptions.protocols.graphql_transport_ws.types import (
    CompleteMessage,
    ConnectionAckMessage,
    ConnectionInitMessage,
    ErrorMessage,
    NextMessage,
    PingMessage,
    PongMessage,
    SubscribeMessage,
)
from tests.http.clients.base import DebuggableGraphQLTransportWSHandler
from tests.views.schema import MyExtension, Schema, Subscription, schema

if TYPE_CHECKING:
    from tests.http.clients.base import HttpClient, WebSocketClient


@pytest_asyncio.fixture
async def ws_raw(http_client: HttpClient) -> AsyncGenerator[WebSocketClient, None]:
    async with http_client.ws_connect(
        "/graphql", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
    ) as ws:
        yield ws
        await ws.close()
        assert ws.closed


@pytest_asyncio.fixture
async def ws(ws_raw: WebSocketClient) -> WebSocketClient:
    await ws_raw.send_message({"type": "connection_init"})
    connection_ack_message: ConnectionAckMessage = await ws_raw.receive_json()
    assert connection_ack_message == {"type": "connection_ack"}
    return ws_raw


def assert_next(
    next_message: NextMessage,
    id: str,
    data: dict[str, object],
    extensions: dict[str, object] | None = None,
):
    """
    Assert that the NextMessage payload contains the provided data.
    If extensions is provided, it will also assert that the
    extensions are present
    """
    assert next_message["type"] == "next"
    assert next_message["id"] == id
    assert set(next_message["payload"].keys()) <= {"data", "errors", "extensions"}
    assert "data" in next_message["payload"]
    assert next_message["payload"]["data"] == data
    if extensions is not None:
        assert "extensions" in next_message["payload"]
        assert next_message["payload"]["extensions"] == extensions


async def test_unknown_message_type(ws_raw: WebSocketClient):
    ws = ws_raw

    await ws.send_json({"type": "NOT_A_MESSAGE_TYPE"})

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == "Unknown message type: NOT_A_MESSAGE_TYPE"


async def test_missing_message_type(ws_raw: WebSocketClient):
    ws = ws_raw

    await ws.send_json({"notType": None})

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == "Failed to parse message"


async def test_parsing_an_invalid_message(ws: WebSocketClient):
    await ws.send_json({"type": "subscribe", "notPayload": None})

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == "Failed to parse message"


async def test_non_text_ws_messages_result_in_socket_closure(ws_raw: WebSocketClient):
    ws = ws_raw

    await ws.send_bytes(
        json.dumps(ConnectionInitMessage({"type": "connection_init"})).encode()
    )

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == "WebSocket message type must be text"


async def test_non_json_ws_messages_result_in_socket_closure(ws_raw: WebSocketClient):
    ws = ws_raw

    await ws.send_text("not valid json")

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == "WebSocket message must be valid JSON"


async def test_ws_message_frame_types_cannot_be_mixed(ws_raw: WebSocketClient):
    ws = ws_raw

    await ws.send_message({"type": "connection_init"})

    ack_message: ConnectionAckMessage = await ws.receive_json()
    assert ack_message == {"type": "connection_ack"}

    await ws.send_bytes(
        json.dumps(
            SubscribeMessage(
                {
                    "id": "sub1",
                    "type": "subscribe",
                    "payload": {
                        "query": "subscription { debug { isConnectionInitTimeoutTaskDone } }"
                    },
                }
            )
        ).encode()
    )

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == "WebSocket message type must be text"


async def test_connection_init_timeout(http_client_class: type[HttpClient]):
    test_client = http_client_class(
        schema, connection_init_wait_timeout=timedelta(seconds=0)
    )

    async with test_client.ws_connect(
        "/graphql", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
    ) as ws:
        await ws.receive(timeout=2)
        assert ws.closed
        assert ws.close_code == 4408
        assert ws.close_reason == "Connection initialisation timeout"


@pytest.mark.flaky
async def test_connection_init_timeout_cancellation(
    ws_raw: WebSocketClient,
):
    # Verify that the timeout task is cancelled after the connection Init
    # message is received
    ws = ws_raw
    await ws.send_message({"type": "connection_init"})

    connection_ack_message: ConnectionAckMessage = await ws.receive_json()
    assert connection_ack_message == {"type": "connection_ack"}

    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {
                "query": "subscription { debug { isConnectionInitTimeoutTaskDone } }"
            },
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert_next(
        next_message, "sub1", {"debug": {"isConnectionInitTimeoutTaskDone": True}}
    )


@pytest.mark.xfail(reason="This test is flaky")
async def test_close_twice(mocker: MockerFixture, http_client_class: type[HttpClient]):
    test_client = http_client_class(
        schema, connection_init_wait_timeout=timedelta(seconds=0.25)
    )

    async with test_client.ws_connect(
        "/graphql", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
    ) as ws:
        transport_close = mocker.patch.object(ws, "close")

        # We set payload is set to "invalid value" to force a invalid payload error
        # which will close the connection
        await ws.send_json({"type": "connection_init", "payload": "invalid value"})

        # Yield control so that ._close can be called
        await asyncio.sleep(0)

        for t in asyncio.all_tasks():
            if (
                t.get_coro().__qualname__
                == "BaseGraphQLTransportWSHandler.handle_connection_init_timeout"
            ):
                # The init timeout task should be cancelled
                with pytest.raises(asyncio.CancelledError):
                    await t

        await ws.receive(timeout=0.5)
        assert ws.closed
        assert ws.close_code == 4400
        assert ws.close_reason == "Invalid connection init payload"
        transport_close.assert_not_called()


async def test_too_many_initialisation_requests(ws: WebSocketClient):
    await ws.send_message({"type": "connection_init"})
    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4429
    assert ws.close_reason == "Too many initialisation requests"


async def test_connections_are_accepted_by_default(ws_raw: WebSocketClient):
    await ws_raw.send_message({"type": "connection_init"})
    connection_ack_message: ConnectionAckMessage = await ws_raw.receive_json()
    assert connection_ack_message == {"type": "connection_ack"}

    await ws_raw.close()
    assert ws_raw.closed


@pytest.mark.parametrize("payload", [None, {"token": "secret"}])
async def test_setting_a_connection_ack_payload(ws_raw: WebSocketClient, payload):
    await ws_raw.send_message(
        {
            "type": "connection_init",
            "payload": {"test-accept": True, "ack-payload": payload},
        }
    )

    connection_ack_message: ConnectionAckMessage = await ws_raw.receive_json()
    assert connection_ack_message == {"type": "connection_ack", "payload": payload}

    await ws_raw.close()
    assert ws_raw.closed


async def test_connection_ack_payload_may_be_unset(ws_raw: WebSocketClient):
    await ws_raw.send_message(
        {
            "type": "connection_init",
            "payload": {"test-accept": True},
        }
    )

    connection_ack_message: ConnectionAckMessage = await ws_raw.receive_json()
    assert connection_ack_message == {"type": "connection_ack"}

    await ws_raw.close()
    assert ws_raw.closed


async def test_rejecting_connection_closes_socket_with_expected_code_and_message(
    ws_raw: WebSocketClient,
):
    await ws_raw.send_message(
        {"type": "connection_init", "payload": {"test-reject": True}}
    )

    await ws_raw.receive(timeout=2)
    assert ws_raw.closed
    assert ws_raw.close_code == 4403
    assert ws_raw.close_reason == "Forbidden"


async def test_context_can_be_modified_from_within_on_ws_connect(
    ws_raw: WebSocketClient,
):
    await ws_raw.send_message(
        {
            "type": "connection_init",
            "payload": {"test-modify": True},
        }
    )

    connection_ack_message: ConnectionAckMessage = await ws_raw.receive_json()
    assert connection_ack_message == {"type": "connection_ack"}

    await ws_raw.send_message(
        {
            "type": "subscribe",
            "id": "demo",
            "payload": {
                "query": "subscription { connectionParams }",
            },
        }
    )

    next_message: NextMessage = await ws_raw.receive_json()
    assert next_message["type"] == "next"
    assert next_message["id"] == "demo"
    assert "data" in next_message["payload"]
    assert next_message["payload"]["data"] == {
        "connectionParams": {"test-modify": True, "modified": True}
    }

    await ws_raw.close()
    assert ws_raw.closed


async def test_ping_pong(ws: WebSocketClient):
    await ws.send_message({"type": "ping"})
    pong_message: PongMessage = await ws.receive_json()
    assert pong_message == {"type": "pong"}


async def test_can_send_payload_with_additional_things(ws_raw: WebSocketClient):
    ws = ws_raw

    # send  init

    await ws.send_message({"type": "connection_init"})

    await ws.receive(timeout=2)

    await ws.send_message(
        {
            "type": "subscribe",
            "payload": {
                "query": 'subscription { echo(message: "Hi") }',
                "extensions": {
                    "some": "other thing",
                },
            },
            "id": "1",
        }
    )

    next_message: NextMessage = await ws.receive_json(timeout=2)

    assert next_message == {
        "type": "next",
        "id": "1",
        "payload": {"data": {"echo": "Hi"}, "extensions": {"example": "example"}},
    }


async def test_server_sent_ping(ws: WebSocketClient):
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": "subscription { requestPing }"},
        }
    )

    ping_message: PingMessage = await ws.receive_json()
    assert ping_message == {"type": "ping"}

    await ws.send_message({"type": "pong"})

    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub1", {"requestPing": True})

    complete_message: CompleteMessage = await ws.receive_json()
    assert complete_message == {"id": "sub1", "type": "complete"}


async def test_unauthorized_subscriptions(ws_raw: WebSocketClient):
    ws = ws_raw
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'subscription { echo(message: "Hi") }'},
        }
    )

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4401
    assert ws.close_reason == "Unauthorized"


async def test_duplicated_operation_ids(ws: WebSocketClient):
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'subscription { echo(message: "Hi", delay: 5) }'},
        }
    )

    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'subscription { echo(message: "Hi", delay: 5) }'},
        }
    )

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4409
    assert ws.close_reason == "Subscriber for sub1 already exists"


async def test_reused_operation_ids(ws: WebSocketClient):
    """Test that an operation id can be re-used after it has been
    previously used for a completed operation.
    """
    # Use sub1 as an id for an operation
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'subscription { echo(message: "Hi") }'},
        }
    )

    next_message1: NextMessage = await ws.receive_json()
    assert_next(next_message1, "sub1", {"echo": "Hi"})

    complete_message: CompleteMessage = await ws.receive_json()
    assert complete_message == {"id": "sub1", "type": "complete"}

    # operation is now complete.  Create a new operation using
    # the same ID
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'subscription { echo(message: "Hi") }'},
        }
    )

    next_message2: NextMessage = await ws.receive_json()
    assert_next(next_message2, "sub1", {"echo": "Hi"})


async def test_simple_subscription(ws: WebSocketClient):
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'subscription { echo(message: "Hi") }'},
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub1", {"echo": "Hi"})
    await ws.send_message({"id": "sub1", "type": "complete"})


@pytest.mark.parametrize(
    ("extra_payload", "expected_message"),
    [
        ({}, "Hi1"),
        ({"operationName": None}, "Hi1"),
        ({"operationName": "Subscription1"}, "Hi1"),
        ({"operationName": "Subscription2"}, "Hi2"),
    ],
)
async def test_operation_selection(
    ws: WebSocketClient, extra_payload, expected_message
):
    await ws.send_json(
        {
            "type": "subscribe",
            "id": "sub1",
            "payload": {
                "query": """
                    subscription Subscription1 { echo(message: "Hi1") }
                    subscription Subscription2 { echo(message: "Hi2") }
                """,
                **extra_payload,
            },
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub1", {"echo": expected_message})
    await ws.send_message({"id": "sub1", "type": "complete"})


@pytest.mark.parametrize(
    ("operation_name"),
    ["", "Subscription2"],
)
async def test_invalid_operation_selection(ws: WebSocketClient, operation_name):
    await ws.send_message(
        {
            "type": "subscribe",
            "id": "sub1",
            "payload": {
                "query": """
                    subscription Subscription1 { echo(message: "Hi1") }
                """,
                "operationName": f"{operation_name}",
            },
        }
    )

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == f'Unknown operation named "{operation_name}".'


async def test_operation_selection_without_operations(ws: WebSocketClient):
    await ws.send_message(
        {
            "type": "subscribe",
            "id": "sub1",
            "payload": {
                "query": """
                    fragment Fragment1 on Query { __typename }
                """,
            },
        }
    )

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == "Can't get GraphQL operation type"


async def test_subscription_syntax_error(ws: WebSocketClient):
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": "subscription { INVALID_SYNTAX "},
        }
    )

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == "Syntax Error: Expected Name, found <EOF>."


async def test_subscription_field_errors(ws: WebSocketClient):
    process_errors = Mock()
    with patch.object(Schema, "process_errors", process_errors):
        await ws.send_message(
            {
                "id": "sub1",
                "type": "subscribe",
                "payload": {
                    "query": "subscription { notASubscriptionField }",
                },
            }
        )

        error_message: ErrorMessage = await ws.receive_json()
        assert error_message["type"] == "error"
        assert error_message["id"] == "sub1"
        assert len(error_message["payload"]) == 1

        assert "locations" in error_message["payload"][0]
        assert error_message["payload"][0]["locations"] == [{"line": 1, "column": 16}]

        assert "message" in error_message["payload"][0]
        assert (
            error_message["payload"][0]["message"]
            == "Cannot query field 'notASubscriptionField' on type 'Subscription'."
        )

        process_errors.assert_called_once()


async def test_subscription_cancellation(ws: WebSocketClient):
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'subscription { echo(message: "Hi", delay: 99) }'},
        }
    )

    await ws.send_message(
        {
            "id": "sub2",
            "type": "subscribe",
            "payload": {
                "query": "subscription { debug { numActiveResultHandlers } }",
            },
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub2", {"debug": {"numActiveResultHandlers": 2}})

    complete_message: CompleteMessage = await ws.receive_json()
    assert complete_message == {"id": "sub2", "type": "complete"}

    await ws.send_message({"id": "sub1", "type": "complete"})

    await ws.send_message(
        {
            "id": "sub3",
            "type": "subscribe",
            "payload": {
                "query": "subscription { debug { numActiveResultHandlers } }",
            },
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub3", {"debug": {"numActiveResultHandlers": 1}})

    complete_message: CompleteMessage = await ws.receive_json()
    assert complete_message == {"id": "sub3", "type": "complete"}


async def test_subscription_errors(ws: WebSocketClient):
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {
                "query": 'subscription { error(message: "TEST ERR") }',
            },
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert next_message["type"] == "next"
    assert next_message["id"] == "sub1"

    assert "errors" in next_message["payload"]
    payload_errors = next_message["payload"]["errors"]
    assert payload_errors is not None
    assert len(payload_errors) == 1

    assert "path" in payload_errors[0]
    assert payload_errors[0]["path"] == ["error"]

    assert "message" in payload_errors[0]
    assert payload_errors[0]["message"] == "TEST ERR"


async def test_operation_error_no_complete(ws: WebSocketClient):
    """Test that an "error" message is not followed by "complete"."""
    # Since we don't include the operation variables,
    # the subscription will fail immediately.
    # see https://github.com/enisdenjo/graphql-ws/blob/master/PROTOCOL.md#error
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {
                "query": "subscription Foo($bar: String!){ exception(message: $bar) }",
            },
        }
    )

    error_message: ErrorMessage = await ws.receive_json()
    assert error_message["type"] == "error"
    assert error_message["id"] == "sub1"

    # after an "error" message, there should be nothing more
    # sent regarding "sub1", not even a "complete".
    await ws.send_message({"type": "ping"})

    pong_message: PongMessage = await ws.receive_json(timeout=1)
    assert pong_message == {"type": "pong"}


async def test_subscription_exceptions(ws: WebSocketClient):
    process_errors = Mock()
    with patch.object(Schema, "process_errors", process_errors):
        await ws.send_message(
            {
                "id": "sub1",
                "type": "subscribe",
                "payload": {
                    "query": 'subscription { exception(message: "TEST EXC") }',
                },
            }
        )

        next_message: NextMessage = await ws.receive_json()
        assert next_message["type"] == "next"
        assert next_message["id"] == "sub1"
        assert "errors" in next_message["payload"]
        assert next_message["payload"]["errors"] == [{"message": "TEST EXC"}]
        process_errors.assert_called_once()


async def test_single_result_query_operation(ws: WebSocketClient):
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": "query { hello }"},
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub1", {"hello": "Hello world"})

    complete_message: CompleteMessage = await ws.receive_json()
    assert complete_message == {"id": "sub1", "type": "complete"}


async def test_single_result_query_operation_async(ws: WebSocketClient):
    """Test a single result query operation on an
    `async` method in the schema, including an artificial
    async delay.
    """
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'query { asyncHello(name: "Dolly", delay:0.01)}'},
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub1", {"asyncHello": "Hello Dolly"})

    complete_message: CompleteMessage = await ws.receive_json()
    assert complete_message == {"id": "sub1", "type": "complete"}


async def test_single_result_query_operation_overlapped(ws: WebSocketClient):
    """Test that two single result queries can be in flight at the same time,
    just like regular queries.  Start two queries with separate ids. The
    first query has a delay, so we expect the message to the second
    query to be delivered first.
    """
    # first query
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'query { asyncHello(name: "Dolly", delay:1)}'},
        }
    )
    # second query
    await ws.send_message(
        {
            "id": "sub2",
            "type": "subscribe",
            "payload": {"query": 'query { asyncHello(name: "Dolly", delay:0)}'},
        }
    )

    # we expect the message to the second query to arrive first
    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub2", {"asyncHello": "Hello Dolly"})

    complete_message: CompleteMessage = await ws.receive_json()
    assert complete_message == {"id": "sub2", "type": "complete"}


async def test_single_result_mutation_operation(ws: WebSocketClient):
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": "mutation { hello }"},
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub1", {"hello": "strawberry"})

    complete_message: CompleteMessage = await ws.receive_json()
    assert complete_message == {"id": "sub1", "type": "complete"}


@pytest.mark.parametrize(
    ("extra_payload", "expected_message"),
    [
        ({}, "Hello Strawberry1"),
        ({"operationName": None}, "Hello Strawberry1"),
        ({"operationName": "Query1"}, "Hello Strawberry1"),
        ({"operationName": "Query2"}, "Hello Strawberry2"),
    ],
)
async def test_single_result_operation_selection(
    ws: WebSocketClient, extra_payload, expected_message
):
    query = """
        query Query1 {
            hello(name: "Strawberry1")
        }
        query Query2 {
            hello(name: "Strawberry2")
        }
    """

    await ws.send_json(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": query, **extra_payload},
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub1", {"hello": expected_message})

    complete_message: CompleteMessage = await ws.receive_json()
    assert complete_message == {"id": "sub1", "type": "complete"}


@pytest.mark.parametrize(
    "operation_name",
    ["", "Query2"],
)
async def test_single_result_invalid_operation_selection(
    ws: WebSocketClient, operation_name
):
    query = """
        query Query1 {
            hello
        }
    """

    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": query, "operationName": operation_name},
        }
    )

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == f'Unknown operation named "{operation_name}".'


async def test_single_result_operation_selection_without_operations(
    ws: WebSocketClient,
):
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {
                "query": """
                    fragment Fragment1 on Query { __typename }
                """,
            },
        }
    )

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == "Can't get GraphQL operation type"


async def test_single_result_execution_error(ws: WebSocketClient):
    process_errors = Mock()
    with patch.object(Schema, "process_errors", process_errors):
        await ws.send_message(
            {
                "id": "sub1",
                "type": "subscribe",
                "payload": {
                    "query": "query { alwaysFail }",
                },
            }
        )

        next_message: NextMessage = await ws.receive_json()
        assert next_message["type"] == "next"
        assert next_message["id"] == "sub1"

        assert "errors" in next_message["payload"]
        payload_errors = next_message["payload"]["errors"]
        assert payload_errors is not None
        assert len(payload_errors) == 1

        assert "path" in payload_errors[0]
        assert payload_errors[0]["path"] == ["alwaysFail"]

        assert "message" in payload_errors[0]
        assert payload_errors[0]["message"] == "You are not authorized"

        process_errors.assert_called_once()


async def test_single_result_pre_execution_error(ws: WebSocketClient):
    """Test that single-result-operations which raise exceptions
    behave in the same way as streaming operations.
    """
    process_errors = Mock()
    with patch.object(Schema, "process_errors", process_errors):
        await ws.send_message(
            {
                "id": "sub1",
                "type": "subscribe",
                "payload": {
                    "query": "query { IDontExist }",
                },
            }
        )

        error_message: ErrorMessage = await ws.receive_json()
        assert error_message["type"] == "error"
        assert error_message["id"] == "sub1"
        assert len(error_message["payload"]) == 1
        assert "message" in error_message["payload"][0]
        assert (
            error_message["payload"][0]["message"]
            == "Cannot query field 'IDontExist' on type 'Query'."
        )
        process_errors.assert_called_once()


async def test_single_result_duplicate_ids_sub(ws: WebSocketClient):
    """Test that single-result-operations and streaming operations
    share the same ID namespace. Start a regular subscription,
    then issue a single-result operation with same ID and expect an
    error due to already existing ID
    """
    # regular subscription
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'subscription { echo(message: "Hi", delay: 5) }'},
        }
    )
    # single result subscription with duplicate id
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {
                "query": "query { hello }",
            },
        }
    )

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4409
    assert ws.close_reason == "Subscriber for sub1 already exists"


async def test_single_result_duplicate_ids_query(ws: WebSocketClient):
    """Test that single-result-operations don't allow duplicate
    IDs for two asynchronous queries. Issue one async query
    with delay, then another with same id. Expect error.
    """
    # single result subscription 1
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'query { asyncHello(name: "Hi", delay: 5) }'},
        }
    )
    # single result subscription with duplicate id
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {
                "query": "query { hello }",
            },
        }
    )

    # We expect the remote to close the socket due to duplicate ID in use
    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4409
    assert ws.close_reason == "Subscriber for sub1 already exists"


async def test_injects_connection_params(ws_raw: WebSocketClient):
    ws = ws_raw
    await ws.send_message(
        {"type": "connection_init", "payload": {"strawberry": "rocks"}}
    )

    connection_ack_message: ConnectionAckMessage = await ws.receive_json()
    assert connection_ack_message == {"type": "connection_ack"}

    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": "subscription { connectionParams }"},
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub1", {"connectionParams": {"strawberry": "rocks"}})

    await ws.send_message({"id": "sub1", "type": "complete"})


async def test_rejects_connection_params_not_dict(ws_raw: WebSocketClient):
    ws = ws_raw
    await ws.send_json({"type": "connection_init", "payload": "gonna fail"})

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == "Invalid connection init payload"


@pytest.mark.parametrize(
    "payload",
    [[], "invalid value", 1],
)
async def test_rejects_connection_params_with_wrong_type(
    payload: object, ws_raw: WebSocketClient
):
    ws = ws_raw
    await ws.send_json({"type": "connection_init", "payload": payload})

    await ws.receive(timeout=2)
    assert ws.closed
    assert ws.close_code == 4400
    assert ws.close_reason == "Invalid connection init payload"


# timings can sometimes fail currently.  Until this test is rewritten when
# generator based subscriptions are implemented, mark it as flaky
@pytest.mark.xfail(reason="This test is flaky, see comment above")
async def test_subsciption_cancel_finalization_delay(ws: WebSocketClient):
    # Test that when we cancel a subscription, the websocket isn't blocked
    # while some complex finalization takes place.
    delay = 0.1

    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": f"subscription {{ longFinalizer(delay: {delay}) }}"},
        }
    )

    next_message: NextMessage = await ws.receive_json()
    assert_next(next_message, "sub1", {"longFinalizer": "hello"})

    # now cancel the stubscription and send a new query. We expect the message
    # to the new query to arrive immediately, without waiting for the finalizer
    start = time.time()
    await ws.send_message({"id": "sub1", "type": "complete"})
    await ws.send_message(
        {
            "id": "sub2",
            "type": "subscribe",
            "payload": {"query": "query { hello }"},
        }
    )

    while True:
        next_or_complete_message: (
            NextMessage | CompleteMessage
        ) = await ws.receive_json()

        assert next_or_complete_message["type"] in ("next", "complete")

        if next_or_complete_message["id"] == "sub2":
            break

    end = time.time()
    elapsed = end - start
    assert elapsed < delay


async def test_error_handler_for_timeout(http_client: HttpClient):
    """Test that the error handler is called when the timeout
    task encounters an error.
    """
    with contextlib.suppress(ImportError):
        from tests.http.clients.channels import ChannelsHttpClient

        if isinstance(http_client, ChannelsHttpClient):
            pytest.skip("Can't patch on_init for this client")

    if not AsyncMock:
        pytest.skip("Don't have AsyncMock")

    ws = ws_raw
    handler = None
    errorhandler = AsyncMock()

    def on_init(_handler):
        nonlocal handler
        if handler:
            return
        handler = _handler
        # patch the object
        handler.handle_task_exception = errorhandler
        # cause an attribute error in the timeout task
        handler.connection_init_wait_timeout = None

    with patch.object(DebuggableGraphQLTransportWSHandler, "on_init", on_init):
        async with http_client.ws_connect(
            "/graphql", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
        ) as ws:
            await asyncio.sleep(0.01)  # wait for the timeout task to start
            await ws.send_message({"type": "connection_init"})
            connection_ack_message: ConnectionAckMessage = await ws.receive_json()
            assert connection_ack_message == {"type": "connection_ack"}
            await ws.close()

    # the error hander should have been called
    assert handler
    errorhandler.assert_called_once()
    args = errorhandler.call_args
    assert isinstance(args[0][0], AttributeError)
    assert "total_seconds" in str(args[0][0])


async def test_subscription_errors_continue(ws: WebSocketClient):
    """Verify that an ExecutionResult with errors during subscription does not terminate
    the subscription.
    """
    process_errors = Mock()
    with patch.object(Schema, "process_errors", process_errors):
        await ws.send_message(
            {
                "id": "sub1",
                "type": "subscribe",
                "payload": {
                    "query": "subscription { flavorsInvalid }",
                },
            }
        )

        next_message1: NextMessage = await ws.receive_json()
        assert next_message1["type"] == "next"
        assert next_message1["id"] == "sub1"
        assert "data" in next_message1["payload"]
        assert next_message1["payload"]["data"] == {"flavorsInvalid": "VANILLA"}

        next_message2: NextMessage = await ws.receive_json()
        assert next_message2["type"] == "next"
        assert next_message2["id"] == "sub1"
        assert "data" in next_message2["payload"]
        assert next_message2["payload"]["data"] is None
        assert "errors" in next_message2["payload"]
        assert "cannot represent value" in str(next_message2["payload"]["errors"])
        process_errors.assert_called_once()

        next_message3: NextMessage = await ws.receive_json()
        assert next_message3["type"] == "next"
        assert next_message3["id"] == "sub1"
        assert "data" in next_message3["payload"]
        assert next_message3["payload"]["data"] == {"flavorsInvalid": "CHOCOLATE"}

        complete_message: CompleteMessage = await ws.receive_json()
        assert complete_message["type"] == "complete"
        assert complete_message["id"] == "sub1"


@patch.object(MyExtension, MyExtension.get_results.__name__, return_value={})
async def test_no_extensions_results_wont_send_extensions_in_payload(
    mock: Mock, ws: WebSocketClient
):
    await ws.send_message(
        {
            "id": "sub1",
            "type": "subscribe",
            "payload": {"query": 'subscription { echo(message: "Hi") }'},
        }
    )

    next_message: NextMessage = await ws.receive_json()
    mock.assert_called_once()
    assert_next(next_message, "sub1", {"echo": "Hi"})
    assert "extensions" not in next_message["payload"]


async def test_unexpected_client_disconnects_are_gracefully_handled(
    ws: WebSocketClient,
):
    process_errors = Mock()

    with patch.object(Schema, "process_errors", process_errors):
        await ws.send_message(
            {
                "id": "sub1",
                "type": "subscribe",
                "payload": {"query": 'subscription { infinity(message: "Hi") }'},
            }
        )
        await ws.receive(timeout=2)
        assert Subscription.active_infinity_subscriptions == 1

        await ws.close()
        await asyncio.sleep(1)

        assert not process_errors.called
        assert Subscription.active_infinity_subscriptions == 0


async def test_max_subscriptions_per_connection(http_client_class: type[HttpClient]):
    """Test that subscriptions beyond the limit are rejected with an error."""
    test_client = http_client_class(schema, max_subscriptions_per_connection=2)

    async with test_client.ws_connect(
        "/graphql", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
    ) as ws:
        await ws.send_message({"type": "connection_init"})
        connection_ack_message: ConnectionAckMessage = await ws.receive_json()
        assert connection_ack_message == {"type": "connection_ack"}

        # First two subscriptions should succeed
        await ws.send_message(
            {
                "id": "sub1",
                "type": "subscribe",
                "payload": {"query": 'subscription { infinity(message: "Hi") }'},
            }
        )
        next_message: NextMessage = await ws.receive_json()
        assert next_message["type"] == "next"
        assert next_message["id"] == "sub1"

        await ws.send_message(
            {
                "id": "sub2",
                "type": "subscribe",
                "payload": {"query": 'subscription { infinity(message: "Hi") }'},
            }
        )
        next_message = await ws.receive_json()
        assert next_message["type"] == "next"
        assert next_message["id"] == "sub2"

        # Third subscription should be rejected
        await ws.send_message(
            {
                "id": "sub3",
                "type": "subscribe",
                "payload": {"query": 'subscription { infinity(message: "Hi") }'},
            }
        )
        error_message: ErrorMessage = await ws.receive_json()
        assert error_message["type"] == "error"
        assert error_message["id"] == "sub3"
        assert error_message["payload"] == [{"message": "Subscription limit reached"}]

        # Completing one should free a slot
        await ws.send_message({"id": "sub1", "type": "complete"})

        await ws.send_message(
            {
                "id": "sub4",
                "type": "subscribe",
                "payload": {"query": 'subscription { infinity(message: "Hi") }'},
            }
        )
        next_message = await ws.receive_json()
        assert next_message["type"] == "next"
        assert next_message["id"] == "sub4"

        await ws.send_message({"id": "sub2", "type": "complete"})
        await ws.send_message({"id": "sub4", "type": "complete"})
        await ws.close()


async def test_max_subscriptions_per_connection_disabled(
    http_client_class: type[HttpClient],
):
    """Test that setting max_subscriptions_per_connection to None disables the limit."""
    test_client = http_client_class(schema, max_subscriptions_per_connection=None)

    async with test_client.ws_connect(
        "/graphql", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
    ) as ws:
        await ws.send_message({"type": "connection_init"})
        connection_ack_message: ConnectionAckMessage = await ws.receive_json()
        assert connection_ack_message == {"type": "connection_ack"}

        # Should be able to create many subscriptions without limit
        for i in range(5):
            await ws.send_message(
                {
                    "id": f"sub{i}",
                    "type": "subscribe",
                    "payload": {"query": 'subscription { infinity(message: "Hi") }'},
                }
            )
            next_message: NextMessage = await ws.receive_json()
            assert next_message["type"] == "next"
            assert next_message["id"] == f"sub{i}"

        for i in range(5):
            await ws.send_message({"id": f"sub{i}", "type": "complete"})
        await ws.close()


async def test_task_done_callback_fires_for_completed_operations(
    http_client: HttpClient,
):
    """The done callback fires for each completed operation task."""
    handler = None
    callback_count = 0
    all_callbacks_fired = asyncio.Event()
    expected_callbacks = 6  # 1 timeout task + 5 operations

    def on_init(_handler):
        nonlocal handler
        handler = _handler
        original_task_done = _handler._task_done

        def tracking_task_done(task):
            nonlocal callback_count
            original_task_done(task)
            callback_count += 1
            if callback_count >= expected_callbacks:
                all_callbacks_fired.set()

        _handler._task_done = tracking_task_done

    with patch.object(DebuggableGraphQLTransportWSHandler, "on_init", on_init):
        async with http_client.ws_connect(
            "/graphql", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
        ) as ws:
            await ws.send_message({"type": "connection_init"})
            await ws.receive_json()  # ack

            for i in range(5):
                sub_id = f"sub{i}"
                await ws.send_message(
                    {
                        "id": sub_id,
                        "type": "subscribe",
                        "payload": {
                            "query": f'subscription {{ echo(message: "msg{i}") }}'
                        },
                    }
                )
                await ws.receive_json()  # next
                complete = await ws.receive_json()  # complete
                assert complete["type"] == "complete"

            await asyncio.wait_for(all_callbacks_fired.wait(), timeout=2)
            await ws.close()

    assert callback_count == expected_callbacks


async def test_done_callback_fires_for_cancelled_subscription(
    http_client: HttpClient,
):
    """The done callback fires for a cancelled subscription task."""
    handler = None
    cancelled_callback_fired = asyncio.Event()

    def on_init(_handler):
        nonlocal handler
        handler = _handler
        original_task_done = _handler._task_done

        def tracking_task_done(task):
            original_task_done(task)
            if task.cancelled():
                cancelled_callback_fired.set()

        _handler._task_done = tracking_task_done

    with patch.object(DebuggableGraphQLTransportWSHandler, "on_init", on_init):
        async with http_client.ws_connect(
            "/graphql", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
        ) as ws:
            await ws.send_message({"type": "connection_init"})
            await ws.receive_json()  # ack

            await ws.send_message(
                {
                    "id": "sub1",
                    "type": "subscribe",
                    "payload": {
                        "query": 'subscription { echo(message: "Hi", delay: 99) }'
                    },
                }
            )

            await ws.send_message({"id": "sub1", "type": "complete"})

            await asyncio.wait_for(cancelled_callback_fired.wait(), timeout=2)
            await ws.close()


async def test_shutdown_with_multiple_in_flight_operations(
    http_client: HttpClient,
):
    """Closing a connection with in-flight subscriptions cancels all tasks cleanly."""
    handler = None
    shutdown_callbacks = 0
    all_shutdown = asyncio.Event()

    def on_init(_handler):
        nonlocal handler
        handler = _handler
        original_task_done = _handler._task_done

        def tracking_task_done(task):
            nonlocal shutdown_callbacks
            original_task_done(task)
            if task.cancelled():
                shutdown_callbacks += 1
                if shutdown_callbacks >= 3:  # 3 operation tasks
                    all_shutdown.set()

        _handler._task_done = tracking_task_done

    with patch.object(DebuggableGraphQLTransportWSHandler, "on_init", on_init):
        async with http_client.ws_connect(
            "/graphql", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
        ) as ws:
            await ws.send_message({"type": "connection_init"})
            await ws.receive_json()  # ack

            for i in range(3):
                await ws.send_message(
                    {
                        "id": f"sub{i}",
                        "type": "subscribe",
                        "payload": {
                            "query": 'subscription { infinity(message: "Hi") }'
                        },
                    }
                )
                next_message: NextMessage = await ws.receive_json()
                assert next_message["type"] == "next"

            assert handler is not None
            assert len(handler.operations) == 3

            await ws.close()

        await asyncio.wait_for(all_shutdown.wait(), timeout=2)
        # 3 operation tasks + possibly the connection_init_timeout_task
        assert shutdown_callbacks >= 3


async def test_task_done_callback_retrieves_exception(
    http_client: HttpClient,
):
    """The done callback retrieves unhandled exceptions from the timeout task.

    Makes handle_task_exception re-raise so the exception escapes the
    coroutine's except block, then verifies the callback retrieved it.
    """
    handler = None
    callback_fired = asyncio.Event()

    def on_init(_handler):
        nonlocal handler
        handler = _handler
        # Break the timeout attribute to cause an AttributeError
        handler.connection_init_wait_timeout = None

        # Make handle_task_exception re-raise so the exception escapes
        async def reraise(error):
            raise error

        handler.handle_task_exception = reraise

        original_task_done = _handler._task_done

        def tracking_task_done(task):
            original_task_done(task)
            callback_fired.set()

        _handler._task_done = tracking_task_done

    with patch.object(DebuggableGraphQLTransportWSHandler, "on_init", on_init):
        async with http_client.ws_connect(
            "/graphql", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
        ) as ws:
            # Wait for the timeout task to fail and the callback to fire
            await asyncio.wait_for(callback_fired.wait(), timeout=2)

            assert handler is not None
            assert handler.connection_init_timeout_task is not None
            assert handler.connection_init_timeout_task.done()
            assert not handler.connection_init_timeout_task.cancelled()
            # The exception escaped the coroutine and was retrieved by _task_done
            assert isinstance(
                handler.connection_init_timeout_task.exception(), AttributeError
            )

            # Clear the timeout task so shutdown() doesn't re-raise
            # when awaiting the already-failed task
            handler.connection_init_timeout_task = None

            await ws.send_message({"type": "connection_init"})
            connection_ack_message: ConnectionAckMessage = await ws.receive_json()
            assert connection_ack_message == {"type": "connection_ack"}
            await ws.close()


async def test_task_done_callback_on_unhandled_operation_exception(
    http_client: HttpClient,
):
    """The done callback retrieves unhandled exceptions from operation tasks.

    Patches run_operation to raise, bypassing the internal except block,
    and verifies the callback fires and retrieves the exception.
    """
    handler = None
    failed_task_exception: list[BaseException] = []
    callback_fired = asyncio.Event()

    def on_init(_handler):
        nonlocal handler
        handler = _handler
        original_task_done = _handler._task_done

        def tracking_task_done(task):
            original_task_done(task)
            if not task.cancelled() and task.exception() is not None:
                failed_task_exception.append(task.exception())
                callback_fired.set()

        _handler._task_done = tracking_task_done

    with patch.object(DebuggableGraphQLTransportWSHandler, "on_init", on_init):
        async with http_client.ws_connect(
            "/graphql", protocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL]
        ) as ws:
            await ws.send_message({"type": "connection_init"})
            await ws.receive_json()  # ack

            assert handler is not None

            async def failing_run(operation):
                raise RuntimeError("simulated crash")

            handler.run_operation = failing_run

            await ws.send_message(
                {
                    "id": "sub1",
                    "type": "subscribe",
                    "payload": {"query": 'subscription { echo(message: "Hi") }'},
                }
            )

            await asyncio.wait_for(callback_fired.wait(), timeout=2)
            await ws.close()

    assert len(failed_task_exception) == 1
    assert isinstance(failed_task_exception[0], RuntimeError)



import pytest
from pytest_codspeed import BenchmarkFixture

from strawberry.schema.config import StrawberryConfig
from strawberry.schema.types.scalar import DEFAULT_SCALAR_REGISTRY
from strawberry.types.arguments import convert_argument
from strawberry.types.base import StrawberryList


@pytest.mark.parametrize("ntypes", [2**k for k in range(14, 23, 2)])
def test_convert_argument_large_list(benchmark: BenchmarkFixture, ntypes):
    test_value = list(range(ntypes))
    type_ = StrawberryList(int)

    def run():
        result = convert_argument(
            test_value, type_, DEFAULT_SCALAR_REGISTRY, StrawberryConfig()
        )
        assert test_value == result

    benchmark(run)



from collections.abc import Sequence
from typing import Optional, Union

import strawberry
from strawberry.annotation import StrawberryAnnotation
from strawberry.types.base import StrawberryList


def test_basic_list():
    annotation = StrawberryAnnotation(list[str])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type is str

    assert resolved == StrawberryList(of_type=str)
    assert resolved == list[str]


def test_basic_tuple():
    annotation = StrawberryAnnotation(tuple[str])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type is str

    assert resolved == StrawberryList(of_type=str)
    assert resolved == tuple[str]


def test_basic_sequence():
    annotation = StrawberryAnnotation(Sequence[str])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type is str

    assert resolved == StrawberryList(of_type=str)
    assert resolved == Sequence[str]


def test_list_of_optional():
    annotation = StrawberryAnnotation(list[int | None])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type == Optional[int]

    assert resolved == StrawberryList(of_type=Optional[int])
    assert resolved == list[int | None]


def test_sequence_of_optional():
    annotation = StrawberryAnnotation(Sequence[int | None])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type == Optional[int]

    assert resolved == StrawberryList(of_type=Optional[int])
    assert resolved == Sequence[int | None]


def test_tuple_of_optional():
    annotation = StrawberryAnnotation(tuple[int | None])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type == Optional[int]

    assert resolved == StrawberryList(of_type=Optional[int])
    assert resolved == tuple[int | None]


def test_list_of_lists():
    annotation = StrawberryAnnotation(list[list[float]])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type == list[float]

    assert resolved == StrawberryList(of_type=list[float])
    assert resolved == list[list[float]]


def test_sequence_of_sequence():
    annotation = StrawberryAnnotation(Sequence[Sequence[float]])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type == Sequence[float]

    assert resolved == StrawberryList(of_type=Sequence[float])
    assert resolved == Sequence[Sequence[float]]


def test_tuple_of_tuple():
    annotation = StrawberryAnnotation(tuple[tuple[float]])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type == tuple[float]

    assert resolved == StrawberryList(of_type=tuple[float])
    assert resolved == tuple[tuple[float]]


def test_list_of_union():
    @strawberry.type
    class Animal:
        feet: bool

    @strawberry.type
    class Fungus:
        spore: bool

    annotation = StrawberryAnnotation(list[Animal | Fungus])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type == Union[Animal, Fungus]

    assert resolved == StrawberryList(of_type=Union[Animal, Fungus])
    assert resolved == list[Animal | Fungus]


def test_sequence_of_union():
    @strawberry.type
    class Animal:
        feet: bool

    @strawberry.type
    class Fungus:
        spore: bool

    annotation = StrawberryAnnotation(Sequence[Animal | Fungus])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type == Union[Animal, Fungus]

    assert resolved == StrawberryList(of_type=Union[Animal, Fungus])
    assert resolved == Sequence[Animal | Fungus]


def test_list_builtin():
    annotation = StrawberryAnnotation(list[str])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type is str

    assert resolved == StrawberryList(of_type=str)
    assert resolved == list[str]
    assert resolved == list[str]


def test_tuple_builtin():
    annotation = StrawberryAnnotation(tuple[str])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryList)
    assert resolved.of_type is str

    assert resolved == StrawberryList(of_type=str)
    assert resolved == tuple[str]
    assert resolved == tuple[str]



from inline_snapshot import snapshot

from .utils.marks import requires_mypy, requires_pyright, requires_ty, skip_on_windows
from .utils.typecheck import Result, typecheck

pytestmark = [skip_on_windows, requires_pyright, requires_mypy, requires_ty]

CODE = """
import strawberry
from strawberry.types.base import StrawberryOptional, StrawberryList


@strawberry.type
class Fruit:
    name: str


reveal_type(StrawberryOptional(Fruit))
reveal_type(StrawberryList(Fruit))
reveal_type(StrawberryOptional(StrawberryList(Fruit)))
reveal_type(StrawberryList(StrawberryOptional(Fruit)))

reveal_type(StrawberryOptional(str))
reveal_type(StrawberryList(str))
reveal_type(StrawberryOptional(StrawberryList(str)))
reveal_type(StrawberryList(StrawberryOptional(str)))
"""


def test():
    results = typecheck(CODE)

    assert results.pyright == snapshot(
        [
            Result(
                type="information",
                message='Type of "StrawberryOptional(Fruit)" is "StrawberryOptional"',
                line=11,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "StrawberryList(Fruit)" is "StrawberryList"',
                line=12,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "StrawberryOptional(StrawberryList(Fruit))" is "StrawberryOptional"',
                line=13,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "StrawberryList(StrawberryOptional(Fruit))" is "StrawberryList"',
                line=14,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "StrawberryOptional(str)" is "StrawberryOptional"',
                line=16,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "StrawberryList(str)" is "StrawberryList"',
                line=17,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "StrawberryOptional(StrawberryList(str))" is "StrawberryOptional"',
                line=18,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "StrawberryList(StrawberryOptional(str))" is "StrawberryList"',
                line=19,
                column=13,
            ),
        ]
    )
    assert results.mypy == snapshot(
        [
            Result(
                type="note",
                message='Revealed type is "strawberry.types.base.StrawberryOptional"',
                line=11,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.types.base.StrawberryList"',
                line=12,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.types.base.StrawberryOptional"',
                line=13,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.types.base.StrawberryList"',
                line=14,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.types.base.StrawberryOptional"',
                line=16,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.types.base.StrawberryList"',
                line=17,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.types.base.StrawberryOptional"',
                line=18,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "strawberry.types.base.StrawberryList"',
                line=19,
                column=13,
            ),
        ]
    )
    assert results.ty == snapshot(
        [
            Result(
                type="information",
                message="Revealed type: `StrawberryOptional`",
                line=11,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `StrawberryList`",
                line=12,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `StrawberryOptional`",
                line=13,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `StrawberryList`",
                line=14,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `StrawberryOptional`",
                line=16,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `StrawberryList`",
                line=17,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `StrawberryOptional`",
                line=18,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `StrawberryList`",
                line=19,
                column=13,
            ),
        ]
    )



import strawberry


def test_type():
    @strawberry.federation.type(keys=["id"])
    class Location:
        id: strawberry.ID

    assert Location(id=strawberry.ID("1")).id == "1"


def test_type_and_override():
    @strawberry.federation.type(keys=["id"])
    class Location:
        id: strawberry.ID
        address: str = strawberry.federation.field(override="start")

    location = Location(id=strawberry.ID("1"), address="ABC")

    assert location.id == "1"
    assert location.address == "ABC"


def test_type_and_override_with_resolver():
    @strawberry.federation.type(keys=["id"])
    class Location:
        id: strawberry.ID
        address: str = strawberry.federation.field(
            override="start", resolver=lambda: "ABC"
        )

    location = Location(id=strawberry.ID("1"))

    assert location.id == "1"



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

import pytest

import strawberry
from strawberry.file_uploads import Upload

if TYPE_CHECKING:
    from sanic import Sanic


@strawberry.type
class Query:
    @strawberry.field
    def index(self) -> str:
        return "Hello there"


@strawberry.type
class Mutation:
    @strawberry.mutation
    def file_upload(self, file: Upload) -> str:
        return file.name


@pytest.fixture
def app():
    from sanic import Sanic

    from strawberry.sanic.views import GraphQLView

    sanic_app = Sanic("sanic_testing")

    sanic_app.add_route(
        GraphQLView.as_view(
            schema=strawberry.Schema(query=Query, mutation=Mutation),
            multipart_uploads_enabled=True,
        ),
        "/graphql",
    )

    return sanic_app


def test_file_cast(app: Sanic):
    """Tests that the list of files in a sanic Request gets correctly turned into a dictionary"""
    from sanic.request import File

    from strawberry.sanic import utils

    file_name = "test.txt"

    file_content = b"Hello, there!."
    in_memory_file = BytesIO(file_content)
    in_memory_file.name = file_name

    form_data = {
        "operations": '{ "query": "mutation($file: Upload!){ fileUpload(file: $file) }", "variables": { "file": null } }',
        "map": '{ "file": ["variables.file"] }',
    }

    files = {
        "file": in_memory_file,
    }

    request, _ = app.test_client.post("/graphql", data=form_data, files=files)

    files = utils.convert_request_to_files_dict(request)  # type: ignore
    file = files["file"]

    assert isinstance(file, File)
    assert file.name == file_name
    assert file.body == file_content


def test_endpoint(app: Sanic):
    """Tests that the graphql api correctly handles file upload and processing"""
    file_name = "test.txt"

    file_content = b"Hello, there!"
    in_memory_file = BytesIO(file_content)
    in_memory_file.name = file_name

    form_data = {
        "operations": '{ "query": "mutation($file: Upload!){ fileUpload(file: $file) }", "variables": { "file": null } }',
        "map": '{ "file": ["variables.file"] }',
    }

    files = {
        "file": in_memory_file,
    }

    _, response = app.test_client.post("/graphql", data=form_data, files=files)

    assert response.json["data"]["fileUpload"] == file_name  # type: ignore



from inline_snapshot import snapshot

from .utils.marks import requires_mypy, requires_pyright, requires_ty, skip_on_windows
from .utils.typecheck import Result, typecheck

pytestmark = [skip_on_windows, requires_pyright, requires_mypy, requires_ty]


CODE = """
import strawberry


@strawberry.federation.type(name="User")
class UserModel:
    name: str


UserModel(name="Patrick")
UserModel(n="Patrick")
"""


def test():
    results = typecheck(CODE)

    assert results.pyright == snapshot(
        [
            Result(
                type="error",
                message='Argument missing for parameter "name"',
                line=11,
                column=1,
            ),
            Result(type="error", message='No parameter named "n"', line=11, column=11),
        ]
    )
    assert results.mypy == snapshot(
        [
            Result(
                type="error",
                message='Unexpected keyword argument "n" for "UserModel"',
                line=11,
                column=1,
            )
        ]
    )
    assert results.ty == snapshot(
        [
            Result(
                type="error",
                message="No argument provided for required parameter `name`",
                line=11,
                column=1,
            ),
            Result(
                type="error",
                message="Argument `n` does not match any known parameter",
                line=11,
                column=11,
            ),
        ]
    )



from graphql import OperationType, parse

from strawberry.utils.operation import get_first_operation


def test_document_without_operation_definition_nodes():
    document = parse(
        """
        fragment Test on Query {
            hello
        }
        """
    )
    assert get_first_operation(document) is None


def test_single_operation_definition_node():
    document = parse(
        """
        query Operation1 {
            hello
        }
        """
    )
    node = get_first_operation(document)
    assert node is not None
    assert node.operation == OperationType.QUERY


def test_multiple_operation_definition_nodes():
    document = parse(
        """
        mutation Operation1 {
            hello
        }
        query Operation2 {
            hello
        }
        """
    )
    node = get_first_operation(document)
    assert node is not None
    assert node.operation == OperationType.MUTATION



from enum import Enum

import strawberry


def test_repr_type():
    @strawberry.type
    class MyType:
        s: str
        i: int
        b: bool
        f: float
        id: strawberry.ID

    assert (
        repr(MyType(s="a", i=1, b=True, f=3.2, id="123"))
        == "test_repr_type.<locals>.MyType(s='a', i=1, b=True, f=3.2, id='123')"
    )


def test_repr_enum():
    @strawberry.enum()
    class Test(Enum):
        A = 1
        B = 2
        C = 3

    assert repr(Test(1)) == "<Test.A: 1>"



import asyncio
from typing import Generic, TypeVar

from pytest_codspeed.plugin import BenchmarkFixture

import strawberry

T = TypeVar("T")


@strawberry.input(description="Filter for GraphQL queries")
class GraphQLFilter(Generic[T]):
    """EXTERNAL Filter for GraphQL queries"""

    eq: T | None = None
    in_: list[T] | None = None
    nin: list[T] | None = None
    gt: T | None = None
    gte: T | None = None
    lt: T | None = None
    lte: T | None = None
    contains: T | None = None
    icontains: T | None = None


@strawberry.type
class Author:
    name: str


@strawberry.type
class Book:
    title: str

    @strawberry.field
    async def authors(
        self,
        name: GraphQLFilter[str] | None = None,
    ) -> list[Author]:
        return [Author(name="F. Scott Fitzgerald")]


def get_books():
    return [
        Book(title="The Great Gatsby"),
    ] * 1000


@strawberry.type
class Query:
    books: list[Book] = strawberry.field(resolver=get_books)


schema = strawberry.Schema(query=Query)

query = """{
    books {
        title
        authors(name: {eq: "F. Scott Fitzgerald"}) {
            name
        }
    }
}
"""


def test_execute_generic_input(benchmark: BenchmarkFixture):
    def run():
        coroutine = schema.execute(query)

        return asyncio.run(coroutine)

    result = benchmark(run)

    assert not result.errors



import dataclasses
import textwrap
import types
from typing import Any, ClassVar, ForwardRef, no_type_check

import pytest

import strawberry
from strawberry.exceptions import (
    MissingArgumentsAnnotationsError,
    MissingFieldAnnotationError,
    MissingReturnAnnotationError,
)
from strawberry.parent import Parent
from strawberry.scalars import JSON
from strawberry.types.fields.resolver import (
    INFO_PARAMSPEC,
    Signature,
    StrawberryResolver,
    UncallableResolverError,
)


def test_resolver_as_argument():
    def get_name(self) -> str:
        return "Name"

    @strawberry.type
    class Query:
        name: str = strawberry.field(resolver=get_name)

    definition = Query.__strawberry_definition__

    assert definition.name == "Query"
    assert len(definition.fields) == 1

    assert definition.fields[0].python_name == "name"
    assert definition.fields[0].graphql_name is None
    assert definition.fields[0].type is str
    assert definition.fields[0].base_resolver.wrapped_func == get_name


def test_resolver_fields():
    @strawberry.type
    class Query:
        @strawberry.field
        def name(self) -> str:
            return "Name"

    definition = Query.__strawberry_definition__

    assert definition.name == "Query"
    assert len(definition.fields) == 1

    assert definition.fields[0].python_name == "name"
    assert definition.fields[0].graphql_name is None
    assert definition.fields[0].type is str
    assert definition.fields[0].base_resolver(None) == Query().name()


def test_staticmethod_resolver_fields():
    @strawberry.type
    class Query:
        @strawberry.field
        @staticmethod
        def name() -> str:
            return "Name"

    definition = Query.__strawberry_definition__

    assert definition.name == "Query"
    assert len(definition.fields) == 1

    assert definition.fields[0].python_name == "name"
    assert definition.fields[0].graphql_name is None
    assert definition.fields[0].type is str
    assert definition.fields[0].base_resolver() == Query.name()

    assert Query.name() == "Name"
    assert Query().name() == "Name"


def test_classmethod_resolver_fields():
    @strawberry.type
    class Query:
        my_val: ClassVar[str] = "thingy"

        @strawberry.field
        @classmethod
        def val(cls) -> str:
            return cls.my_val

    definition = Query.__strawberry_definition__

    assert definition.name == "Query"
    assert len(definition.fields) == 1

    assert definition.fields[0].python_name == "val"
    assert definition.fields[0].graphql_name is None
    assert definition.fields[0].type is str
    assert definition.fields[0].base_resolver() == Query.val()

    assert Query.val() == "thingy"
    assert Query().val() == "thingy"


@pytest.mark.raises_strawberry_exception(
    MissingReturnAnnotationError,
    match='Return annotation missing for field "hello", did you forget to add it?',
)
def test_raises_error_when_return_annotation_missing():
    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self):
            return "I'm a resolver"


@pytest.mark.raises_strawberry_exception(
    MissingReturnAnnotationError,
    match='Return annotation missing for field "hello", did you forget to add it?',
)
def test_raises_error_when_return_annotation_missing_async_function():
    @strawberry.type
    class Query:
        @strawberry.field
        async def hello(self):
            return "I'm a resolver"


@pytest.mark.raises_strawberry_exception(
    MissingReturnAnnotationError,
    match='Return annotation missing for field "goodbye", did you forget to add it?',
)
def test_raises_error_when_return_annotation_missing_resolver():
    @strawberry.type
    class Query2:
        def adios(self):
            return -1

        goodbye = strawberry.field(resolver=adios)


@pytest.mark.raises_strawberry_exception(
    MissingArgumentsAnnotationsError,
    match=(
        'Missing annotation for argument "query" in field "hello", '
        "did you forget to add it?"
    ),
)
def test_raises_error_when_argument_annotation_missing():
    @strawberry.field
    def hello(self, query) -> str:
        return "I'm a resolver"


@pytest.mark.raises_strawberry_exception(
    MissingArgumentsAnnotationsError,
    match=(
        'Missing annotation for arguments "query" and "limit" '
        'in field "hello", did you forget to add it?'
    ),
)
def test_raises_error_when_argument_annotation_missing_multiple_fields():
    @strawberry.field
    def hello(self, query, limit) -> str:
        return "I'm a resolver"


@pytest.mark.raises_strawberry_exception(
    MissingArgumentsAnnotationsError,
    match=(
        'Missing annotation for argument "query" '
        'in field "hello", did you forget to add it?'
    ),
)
def test_raises_error_when_argument_annotation_missing_multiple_lines():
    @strawberry.field
    def hello(
        self,
        query,
    ) -> str:
        return "I'm a resolver"


@pytest.mark.raises_strawberry_exception(
    MissingArgumentsAnnotationsError,
    match=(
        'Missing annotation for argument "query" '
        'in field "hello", did you forget to add it?'
    ),
)
def test_raises_error_when_argument_annotation_missing_default_value():
    @strawberry.field
    def hello(
        self,
        query="this is a default value",
    ) -> str:
        return "I'm a resolver"


@pytest.mark.raises_strawberry_exception(
    MissingFieldAnnotationError,
    match=(
        'Unable to determine the type of field "missing". '
        "Either annotate it directly, or provide a typed resolver "
        "using @strawberry.field."
    ),
)
def test_raises_error_when_missing_annotation_and_resolver():
    @strawberry.type
    class Query:
        missing = strawberry.field(name="annotation")


@pytest.mark.raises_strawberry_exception(
    MissingFieldAnnotationError,
    match=(
        'Unable to determine the type of field "missing". Either annotate it '
        "directly, or provide a typed resolver using @strawberry.field."
    ),
)
def test_raises_error_when_missing_type():
    """Test to make sure that if somehow a non-StrawberryField field is added to the cls
    without annotations it raises an exception. This would occur if someone manually
    uses dataclasses.field
    """

    @strawberry.type
    class Query:
        missing = dataclasses.field()


@pytest.mark.raises_strawberry_exception(
    MissingFieldAnnotationError,
    match=(
        'Unable to determine the type of field "missing". Either annotate it '
        "directly, or provide a typed resolver using @strawberry.field."
    ),
)
def test_raises_error_when_missing_type_on_dynamic_class():
    # this test if for making sure the code that finds the exception source
    # doesn't crash with dynamic code

    namespace = {"missing": dataclasses.field()}

    strawberry.type(types.new_class("Query", (), {}, lambda ns: ns.update(namespace)))


@pytest.mark.raises_strawberry_exception(
    MissingFieldAnnotationError,
    match=(
        'Unable to determine the type of field "banana". Either annotate it '
        "directly, or provide a typed resolver using @strawberry.field."
    ),
)
def test_raises_error_when_missing_type_on_longish_class():
    @strawberry.type
    class Query:
        field_1: str = strawberry.field(name="field_1")
        field_2: str = strawberry.field(name="field_2")
        field_3: str = strawberry.field(name="field_3")
        field_4: str = strawberry.field(name="field_4")
        field_5: str = strawberry.field(name="field_5")
        field_6: str = strawberry.field(name="field_6")
        field_7: str = strawberry.field(name="field_7")
        field_8: str = strawberry.field(name="field_8")
        field_9: str = strawberry.field(name="field_9")
        banana = strawberry.field(name="banana")
        field_10: str = strawberry.field(name="field_10")
        field_11: str = strawberry.field(name="field_11")
        field_12: str = strawberry.field(name="field_12")
        field_13: str = strawberry.field(name="field_13")
        field_14: str = strawberry.field(name="field_14")
        field_15: str = strawberry.field(name="field_15")
        field_16: str = strawberry.field(name="field_16")
        field_17: str = strawberry.field(name="field_17")
        field_18: str = strawberry.field(name="field_18")
        field_19: str = strawberry.field(name="field_19")


def test_raises_error_calling_uncallable_resolver():
    @classmethod  # type: ignore
    def class_func(cls) -> int: ...

    # Note that class_func is a raw classmethod object because it has not been bound
    # to a class at this point
    resolver = StrawberryResolver(class_func)

    with pytest.raises(
        UncallableResolverError,
        match=r"Attempted to call resolver (.*) with uncallable function (.*)",
    ):
        resolver()


def test_can_reuse_resolver():
    def get_name(self) -> str:
        return "Name"

    @strawberry.type
    class Query:
        name: str = strawberry.field(resolver=get_name)
        name_2: str = strawberry.field(resolver=get_name)

    definition = Query.__strawberry_definition__

    assert definition.name == "Query"
    assert len(definition.fields) == 2

    assert definition.fields[0].python_name == "name"
    assert definition.fields[0].graphql_name is None
    assert definition.fields[0].python_name == "name"
    assert definition.fields[0].type is str
    assert definition.fields[0].base_resolver.wrapped_func == get_name

    assert definition.fields[1].python_name == "name_2"
    assert definition.fields[1].graphql_name is None
    assert definition.fields[1].python_name == "name_2"
    assert definition.fields[1].type is str
    assert definition.fields[1].base_resolver.wrapped_func == get_name


def test_eq_resolvers():
    def get_name(self) -> str:
        return "Name"

    @strawberry.type
    class Query:
        a: int
        name: str = strawberry.field(resolver=get_name)
        name_2: str = strawberry.field(resolver=get_name)

    assert Query(a=1) == Query(a=1)
    assert Query(a=1) != Query(a=2)


def test_eq_fields():
    @strawberry.type
    class Query:
        a: int
        name: str = strawberry.field(name="name")

    assert Query(a=1, name="name") == Query(a=1, name="name")
    assert Query(a=1, name="name") != Query(a=1, name="not a name")


def test_with_resolver_fields():
    @strawberry.type
    class Query:
        a: int

        @strawberry.field
        def name(self) -> str:
            return "A"

    assert Query(a=1) == Query(a=1)
    assert Query(a=1) != Query(a=2)


def root_and_info(
    root,
    foo: str,
    bar: float,
    info: str,
    strawberry_info: strawberry.Info,
) -> str:
    raise AssertionError("Unreachable code.")


def self_and_info(
    self,
    foo: str,
    bar: float,
    info: str,
    strawberry_info: strawberry.Info,
) -> str:
    raise AssertionError("Unreachable code.")


def parent_and_info(
    parent: Parent[str],
    foo: str,
    bar: float,
    info: str,
    strawberry_info: strawberry.Info,
) -> str:
    raise AssertionError("Unreachable code.")


@pytest.mark.parametrize(
    "resolver_func",
    [
        pytest.param(self_and_info),
        pytest.param(root_and_info),
        pytest.param(parent_and_info),
    ],
)
def test_resolver_annotations(resolver_func):
    """Ensure only non-reserved annotations are returned."""
    resolver = StrawberryResolver(resolver_func)

    expected_annotations = {"foo": str, "bar": float, "info": str, "return": str}
    assert resolver.annotations == expected_annotations

    # Sanity-check to ensure StrawberryArguments return the same annotations
    assert {
        **{
            arg.python_name: arg.type_annotation.resolve()  # type: ignore
            for arg in resolver.arguments
        },
        "return": str,
    }


@no_type_check
def test_resolver_with_unhashable_default():
    @strawberry.type
    class Query:
        @strawberry.field
        def field(self, x: list[str] = ["foo"], y: JSON = {"foo": 42}) -> str:  # noqa: B006
            return f"{x} {y}"

    schema = strawberry.Schema(Query)
    result = schema.execute_sync("query { field }")
    assert result.data == {"field": "['foo'] {'foo': 42}"}
    assert not result.errors


@no_type_check
def test_parameter_hash_collision():
    """Ensure support for hashable defaults does not introduce collision."""

    def foo(x: str = "foo"):
        pass

    def bar(x: str = "bar"):
        pass

    foo_signature = Signature.from_callable(foo, follow_wrapped=True)
    bar_signature = Signature.from_callable(bar, follow_wrapped=True)

    foo_param = foo_signature.parameters["x"]
    bar_param = bar_signature.parameters["x"]

    # Ensure __eq__ still functions properly
    assert foo_param != bar_param

    # Ensure collision does not occur in hash-map and hash-tables. Colisions are
    # prevented by Python invoking __eq__ when two items have the same hash.
    parameters_map = {
        foo_param: "foo",
        bar_param: "bar",
    }
    parameters_set = {foo_param, bar_param}

    assert len(parameters_map) == 2
    assert len(parameters_set) == 2


def test_annotation_using_parent_annotation():
    @strawberry.type
    class FruitType:
        name: str

        @strawberry.field
        @staticmethod
        def name_from_parent(parent: strawberry.Parent[Any]) -> str:
            return f"Using 'parent': {parent.name}"

    @strawberry.type
    class Query:
        @strawberry.field
        @staticmethod
        def fruit() -> FruitType:
            return FruitType(name="Strawberry")

    schema = strawberry.Schema(query=Query)
    expected = """\
      type FruitType {
        name: String!
        nameFromParent: String!
      }

      type Query {
        fruit: FruitType!
      }
    """

    assert textwrap.dedent(str(schema)) == textwrap.dedent(expected).strip()

    result = schema.execute_sync("query { fruit { name nameFromParent } }")
    assert result.data == {
        "fruit": {
            "name": "Strawberry",
            "nameFromParent": "Using 'parent': Strawberry",
        }
    }


def test_annotation_using_parent_annotation_but_named_root():
    @strawberry.type
    class FruitType:
        name: str

        @strawberry.field
        @staticmethod
        def name_from_parent(root: strawberry.Parent[Any]) -> str:
            return f"Using 'root': {root.name}"

    @strawberry.type
    class Query:
        @strawberry.field
        @staticmethod
        def fruit() -> FruitType:
            return FruitType(name="Strawberry")

    schema = strawberry.Schema(query=Query)
    expected = """\
      type FruitType {
        name: String!
        nameFromParent: String!
      }

      type Query {
        fruit: FruitType!
      }
    """

    assert textwrap.dedent(str(schema)) == textwrap.dedent(expected).strip()

    result = schema.execute_sync("query { fruit { name nameFromParent } }")
    assert result.data == {
        "fruit": {
            "name": "Strawberry",
            "nameFromParent": "Using 'root': Strawberry",
        }
    }


def test_info_annotation_with_type_alias():
    """Test that TypeAliasType for Info is properly recognized (issue #4057)."""
    from typing_extensions import TypeAliasType

    from strawberry.types.info import Info

    # Simulate Python 3.12's `type MyInfo = Info[None, None]`
    MyInfo = TypeAliasType("MyInfo", Info[None, None])

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self, info: MyInfo) -> str:
            return "Hello World"

    # Should not raise DeprecationWarning
    schema = strawberry.Schema(query=Query)
    result = schema.execute_sync("query { hello }")
    assert result.errors is None
    assert result.data == {"hello": "Hello World"}


def test_info_forward_reference_string():
    """Test that string forward references for Info are properly recognized.

    Only qualified names (strawberry.Info or full module path) are matched
    to avoid confusion with user-defined types.
    """

    # Test qualified string patterns that should match Info
    assert INFO_PARAMSPEC.is_reserved_type("strawberry.Info") is True
    assert INFO_PARAMSPEC.is_reserved_type("strawberry.Info[Context, None]") is True
    assert INFO_PARAMSPEC.is_reserved_type("strawberry.types.info.Info") is True
    assert (
        INFO_PARAMSPEC.is_reserved_type("strawberry.types.info.Info[Context, None]")
        is True
    )

    # Test patterns that should NOT match (unqualified or wrong names)
    assert INFO_PARAMSPEC.is_reserved_type("Info") is False  # unqualified
    assert INFO_PARAMSPEC.is_reserved_type("Info[Context, None]") is False
    assert INFO_PARAMSPEC.is_reserved_type("str") is False
    assert INFO_PARAMSPEC.is_reserved_type("InfoClass") is False
    assert INFO_PARAMSPEC.is_reserved_type("MyInfo") is False


def test_info_forward_ref_object():
    """Test that ForwardRef objects for Info are properly recognized.

    Only qualified names (strawberry.Info or full module path) are matched
    to avoid confusion with user-defined types.
    """

    from strawberry.types.fields.resolver import INFO_PARAMSPEC

    # Test ForwardRef patterns that should match Info
    assert INFO_PARAMSPEC.is_reserved_type(ForwardRef("strawberry.Info")) is True
    assert (
        INFO_PARAMSPEC.is_reserved_type(ForwardRef("strawberry.Info[Context, None]"))
        is True
    )
    assert (
        INFO_PARAMSPEC.is_reserved_type(ForwardRef("strawberry.types.info.Info"))
        is True
    )

    # Test ForwardRef patterns that should NOT match
    assert INFO_PARAMSPEC.is_reserved_type(ForwardRef("Info")) is False  # unqualified
    assert INFO_PARAMSPEC.is_reserved_type(ForwardRef("str")) is False
    assert INFO_PARAMSPEC.is_reserved_type(ForwardRef("MyInfo")) is False



import textwrap
from typing import Annotated

import strawberry
from strawberry.field_extensions import InputMutationExtension


@strawberry.federation.type
class Fruit:
    name: str
    color: str


@strawberry.federation.type
class Query:
    @strawberry.field
    def hello(self) -> str:  # pragma: no cover
        return "hi"


@strawberry.federation.type
class Mutation:
    @strawberry.federation.mutation(extensions=[InputMutationExtension()])
    def create_fruit(
        self,
        name: str,
        color: Annotated[
            str,
            strawberry.federation.argument(
                description="The color of the fruit",
            ),
        ],
    ) -> Fruit:
        return Fruit(
            name=name,
            color=color,
        )


schema = strawberry.federation.Schema(query=Query, mutation=Mutation)


def test_schema():
    expected = '''
    input CreateFruitInput {
      name: String!

      """The color of the fruit"""
      color: String!
    }

    type Fruit {
      name: String!
      color: String!
    }

    type Mutation {
      createFruit(
        """Input data for `createFruit` mutation"""
        input: CreateFruitInput!
      ): Fruit!
    }

    type Query {
      _service: _Service!
      hello: String!
    }

    scalar _Any

    type _Service {
      sdl: String!
    }
    '''
    assert str(schema).strip() == textwrap.dedent(expected).strip()


def test_input_mutation():
    result = schema.execute_sync(
        """
        mutation TestQuery ($input: CreateFruitInput!) {
            createFruit (input: $input) {
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "input": {
                "name": "Dragonfruit",
                "color": "red",
            }
        },
    )
    assert result.errors is None
    assert result.data == {
        "createFruit": {
            "name": "Dragonfruit",
            "color": "red",
        },
    }



import os
import subprocess
import sys
import sysconfig
import textwrap
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Generic, TypeVar

import pytest

import strawberry

if TYPE_CHECKING:
    from tests.schema.test_lazy.type_a import TypeA
    from tests.schema.test_lazy.type_c import TypeC

STRAWBERRY_EXECUTABLE = next(
    Path(sysconfig.get_path("scripts")).glob("strawberry*"), None
)

T = TypeVar("T")

TypeAType = Annotated["TypeA", strawberry.lazy("tests.schema.test_lazy.type_a")]


def test_lazy_types_with_generic():
    @strawberry.type
    class Edge(Generic[T]):
        node: T

    @strawberry.type
    class Query:
        users: Edge[TypeAType]

    strawberry.Schema(query=Query)


def test_no_generic_type_duplication_with_lazy():
    from tests.schema.test_lazy.type_b import TypeB

    @strawberry.type
    class Edge(Generic[T]):
        node: T

    @strawberry.type
    class Query:
        users: Edge[TypeB]
        relatively_lazy_users: Edge[Annotated["TypeB", strawberry.lazy(".type_b")]]
        absolutely_lazy_users: Edge[
            Annotated["TypeB", strawberry.lazy("tests.schema.test_lazy.type_b")]
        ]

    schema = strawberry.Schema(query=Query)

    expected_schema = textwrap.dedent(
        """
        type Query {
          users: TypeBEdge!
          relativelyLazyUsers: TypeBEdge!
          absolutelyLazyUsers: TypeBEdge!
        }

        type TypeA {
          listOfB: [TypeB!]
          typeB: TypeB!
        }

        type TypeB {
          typeA: TypeA!
          typeAList: [TypeA!]!
          typeCList: [TypeC!]!
        }

        type TypeBEdge {
          node: TypeB!
        }

        type TypeC {
          name: String!
        }
        """
    ).strip()

    assert str(schema) == expected_schema


@pytest.mark.parametrize(
    "commands",
    [
        pytest.param(
            [sys.executable, "tests/schema/test_lazy/type_c.py"],
            id="script",
        ),
        pytest.param(
            [sys.executable, "-m", "tests.schema.test_lazy.type_c"],
            id="module",
        ),
        pytest.param(
            [STRAWBERRY_EXECUTABLE, "export-schema", "tests.schema.test_lazy.schema"],
            id="cli",
            marks=pytest.mark.skipif(
                sys.platform == "win32", reason="Test is broken on windows"
            ),
        ),
    ],
)
def test_lazy_types_loaded_from_same_module(commands: Sequence[str]):
    """Test if lazy types resolved from the same module produce duplication error.

    Note:
      `subprocess` is used since the test must be run as the main module / script.
    """
    result = subprocess.run(
        args=[*commands],
        env=os.environ,
        capture_output=True,
        check=True,
    )

    expected = """\
    type Query {
      typeA: TypeCEdge!
      typeB: TypeCEdge!
    }

    type TypeC {
      name: String!
    }

    type TypeCEdge {
      node: TypeC!
    }
    """

    schema_sdl = result.stdout.decode().replace(os.linesep, "\n")
    assert textwrap.dedent(schema_sdl) == textwrap.dedent(expected)


def test_lazy_types_declared_within_optional():
    from tests.schema.test_lazy.type_c import Edge, TypeC

    @strawberry.type
    class Query:
        normal_edges: list[Edge[TypeC | None]]
        lazy_edges: list[
            Edge[
                Annotated["TypeC", strawberry.lazy("tests.schema.test_lazy.type_c")]
                | None
            ]
        ]

    schema = strawberry.Schema(query=Query)
    expected_schema = textwrap.dedent(
        """
        type Query {
          normalEdges: [TypeCOptionalEdge!]!
          lazyEdges: [TypeCOptionalEdge!]!
        }

        type TypeC {
          name: String!
        }

        type TypeCOptionalEdge {
          node: TypeC
        }
        """
    ).strip()

    assert str(schema) == expected_schema


def test_lazy_with_already_specialized_generic():
    from tests.schema.test_lazy.type_d import Query

    schema = strawberry.Schema(query=Query)
    expected_schema = textwrap.dedent(
        """
        type Query {
          typeD1: TypeD!
          typeD: TypeD!
        }

        type TypeD {
          name: String!
        }
        """
    ).strip()

    assert str(schema) == expected_schema



from typing import TYPE_CHECKING, Annotated

import strawberry

if TYPE_CHECKING:
    from .type_a import TypeA


@strawberry.type
class TypeB:
    @strawberry.field()
    def type_a(
        self,
    ) -> Annotated["TypeA", strawberry.lazy("tests.schema.test_lazy_types.type_a")]:
        from .type_a import TypeA

        return TypeA()



import textwrap

from strawberry.schema_codegen import codegen


def test_generates_used_interface_before():
    schema = """
    type Human implements Being {
        id: ID!
        name: String!
        friends: [Human]
    }

    type Cat implements Being {
        id: ID!
        name: String!
        livesLeft: Int
    }

    interface Being {
        id: ID!
        name: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.interface
        class Being:
            id: strawberry.ID
            name: str

        @strawberry.type
        class Human(Being):
            id: strawberry.ID
            name: str
            friends: list[Human | None] | None

        @strawberry.type
        class Cat(Being):
            id: strawberry.ID
            name: str
            lives_left: int | None
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_forward_reference_in_field_annotation():
    """A type can reference another type that appears later in the generated output.

    The generated ``from __future__ import annotations`` (PEP 563) defers
    evaluation of annotations, so ``foo: Foo`` does not require ``Foo`` to be
    defined at class-creation time.
    """
    import subprocess
    import sys
    import tempfile
    from pathlib import Path

    schema = """
    type FooContainer {
        foo: Foo!
    }

    type Foo {
        name: String!
    }
    """
    code = codegen(schema)
    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as f:
        f.write(code)
        path = f.name
    try:
        result = subprocess.run(
            [sys.executable, "-c", f"import runpy; runpy.run_path({path!r})"],
            capture_output=True,
            text=True,
            timeout=5,
            check=False,
        )
        assert result.returncode == 0, (
            f"Generated module should import without error; stderr: {result.stderr}"
        )
    finally:
        Path(path).unlink()


def test_generates_union_members_before_union():
    """Union member types must be emitted before the union definition.

    The generated union assignment (e.g.
    ``FooOrBar = Annotated[Foo | Bar, strawberry.union(...)]``) is a runtime
    expression, not an annotation, so ``from __future__ import annotations``
    would not defer its evaluation.  The member types must therefore be defined
    before the union line.
    """
    schema = """
    interface Base {
        id: ID!
    }

    type Foo implements Base {
        id: ID!
        name: String!
    }

    type Bar implements Base {
        id: ID!
        title: String!
    }

    union FooOrBar = Foo | Bar
    """

    generated = codegen(schema).strip()

    base_pos = generated.index("@strawberry.interface\nclass Base:")
    foo_pos = generated.index("@strawberry.type\nclass Foo(Base):")
    bar_pos = generated.index("@strawberry.type\nclass Bar(Base):")
    union_pos = generated.index("FooOrBar = Annotated[")

    assert base_pos < foo_pos, "Interface must precede implementing type"
    assert base_pos < bar_pos, "Interface must precede implementing type"
    assert foo_pos < union_pos, "Union members must precede union definition"
    assert bar_pos < union_pos, "Union members must precede union definition"



import sys
from pathlib import Path

import pytest

from strawberry.exceptions.exception_source import ExceptionSource

pytestmark = pytest.mark.skipif(
    sys.platform == "win32", reason="Test is meant to run on Unix systems"
)


def test_returns_relative_path(mocker):
    mocker.patch.object(Path, "cwd", return_value="/home/user/project/")

    source = ExceptionSource(
        path=Path("/home/user/project/src/main.py"),
        code="",
        start_line=1,
        end_line=1,
        error_line=1,
        error_column=1,
        error_column_end=1,
    )

    assert source.path_relative_to_cwd == Path("src/main.py")


def test_returns_relative_path_when_is_already_relative():
    source = ExceptionSource(
        path=Path("src/main.py"),
        code="",
        start_line=1,
        end_line=1,
        error_line=1,
        error_column=1,
        error_column_end=1,
    )

    assert source.path_relative_to_cwd == Path("src/main.py")



from pathlib import Path

from inline_snapshot import snapshot

from strawberry.utils.importer import import_module_symbol
from strawberry.utils.locate_definition import locate_definition
from tests.typecheckers.utils.marks import skip_on_windows

pytestmark = skip_on_windows


def _simplify_path(path: str) -> str:
    path = Path(path)

    root = Path(__file__).parents[1]

    return str(path.relative_to(root))


def test_find_model_name() -> None:
    schema = import_module_symbol(
        "tests.fixtures.sample_package.sample_module", default_symbol_name="schema"
    )
    result = locate_definition(schema, "User")

    assert _simplify_path(result) == snapshot(
        "fixtures/sample_package/sample_module.py:38:7"
    )


def test_find_model_name_enum() -> None:
    schema = import_module_symbol(
        "tests.fixtures.sample_package.sample_module", default_symbol_name="schema"
    )
    result = locate_definition(schema, "Role")

    assert _simplify_path(result) == snapshot(
        "fixtures/sample_package/sample_module.py:32:7"
    )


def test_find_model_name_scalar() -> None:
    schema = import_module_symbol(
        "tests.fixtures.sample_package.sample_module", default_symbol_name="schema"
    )
    result = locate_definition(schema, "ExampleScalar")

    assert _simplify_path(result) == snapshot(
        "fixtures/sample_package/sample_module.py:7:13"
    )


def test_find_model_field() -> None:
    schema = import_module_symbol(
        "tests.fixtures.sample_package.sample_module", default_symbol_name="schema"
    )
    result = locate_definition(schema, "User.name")

    assert _simplify_path(result) == snapshot(
        "fixtures/sample_package/sample_module.py:39:5"
    )


def test_find_model_field_scalar() -> None:
    schema = import_module_symbol(
        "tests.fixtures.sample_package.sample_module", default_symbol_name="schema"
    )
    result = locate_definition(schema, "User.example_scalar")

    assert _simplify_path(result) == snapshot(
        "fixtures/sample_package/sample_module.py:42:5"
    )


def test_find_model_field_with_resolver() -> None:
    schema = import_module_symbol(
        "tests.fixtures.sample_package.sample_module", default_symbol_name="schema"
    )
    result = locate_definition(schema, "Query.user")

    assert _simplify_path(result) == snapshot(
        "fixtures/sample_package/sample_module.py:50:5"
    )


def test_find_missing_model() -> None:
    schema = import_module_symbol(
        "tests.fixtures.sample_package.sample_module", default_symbol_name="schema"
    )
    result = locate_definition(schema, "Missing")

    assert result is None


def test_find_missing_model_field() -> None:
    schema = import_module_symbol(
        "tests.fixtures.sample_package.sample_module", default_symbol_name="schema"
    )
    result = locate_definition(schema, "Missing.field")

    assert result is None


def test_find_union() -> None:
    schema = import_module_symbol(
        "tests.fixtures.sample_package.sample_module", default_symbol_name="schema"
    )
    result = locate_definition(schema, "UnionExample")

    assert _simplify_path(result) == snapshot(
        "fixtures/sample_package/sample_module.py:23:16"
    )


def test_find_inline_union() -> None:
    schema = import_module_symbol(
        "tests.fixtures.sample_package.sample_module", default_symbol_name="schema"
    )
    result = locate_definition(schema, "InlineUnion")

    assert _simplify_path(result) == snapshot(
        "fixtures/sample_package/sample_module.py:44:19"
    )



import textwrap
from enum import Enum
from typing import Annotated

import strawberry


def test_field_tag_printed_correctly():
    @strawberry.federation.interface(tags=["myTag", "anotherTag"])
    class SomeInterface:
        id: strawberry.ID

    @strawberry.federation.type(tags=["myTag", "anotherTag"])
    class Product(SomeInterface):
        upc: str = strawberry.federation.field(
            external=True, tags=["myTag", "anotherTag"]
        )

    @strawberry.federation.type
    class Query:
        @strawberry.field
        def top_products(
            self, first: Annotated[int, strawberry.federation.argument(tags=["myTag"])]
        ) -> list[Product]:  # pragma: no cover
            return []

    schema = strawberry.federation.Schema(query=Query)

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@external", "@tag"]) {
          query: Query
        }

        type Product implements SomeInterface @tag(name: "myTag") @tag(name: "anotherTag") {
          id: ID!
          upc: String! @external @tag(name: "myTag") @tag(name: "anotherTag")
        }

        type Query {
          _service: _Service!
          topProducts(first: Int! @tag(name: "myTag")): [Product!]!
        }

        interface SomeInterface @tag(name: "myTag") @tag(name: "anotherTag") {
          id: ID!
        }

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()


def test_field_tag_printed_correctly_on_scalar():
    @strawberry.federation.scalar(tags=["myTag", "anotherTag"])
    class SomeScalar(str):
        __slots__ = ()

    @strawberry.federation.type
    class Query:
        hello: SomeScalar

    schema = strawberry.federation.Schema(query=Query)

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@tag"]) {
          query: Query
        }

        type Query {
          _service: _Service!
          hello: SomeScalar!
        }

        scalar SomeScalar @tag(name: "myTag") @tag(name: "anotherTag")

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()


def test_field_tag_printed_correctly_on_enum():
    @strawberry.federation.enum(tags=["myTag", "anotherTag"])
    class SomeEnum(Enum):
        A = "A"

    @strawberry.federation.type
    class Query:
        hello: SomeEnum

    schema = strawberry.federation.Schema(query=Query)

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@tag"]) {
          query: Query
        }

        type Query {
          _service: _Service!
          hello: SomeEnum!
        }

        enum SomeEnum @tag(name: "myTag") @tag(name: "anotherTag") {
          A
        }

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()


def test_field_tag_printed_correctly_on_enum_value():
    @strawberry.enum
    class SomeEnum(Enum):
        A = strawberry.federation.enum_value("A", tags=["myTag", "anotherTag"])

    @strawberry.federation.type
    class Query:
        hello: SomeEnum

    schema = strawberry.federation.Schema(query=Query)

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@tag"]) {
          query: Query
        }

        type Query {
          _service: _Service!
          hello: SomeEnum!
        }

        enum SomeEnum {
          A @tag(name: "myTag") @tag(name: "anotherTag")
        }

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()


def test_field_tag_printed_correctly_on_union():
    @strawberry.type
    class A:
        a: str

    @strawberry.type
    class B:
        b: str

    MyUnion = Annotated[
        A | B, strawberry.federation.union("Union", tags=["myTag", "anotherTag"])
    ]

    @strawberry.federation.type
    class Query:
        hello: MyUnion

    schema = strawberry.federation.Schema(query=Query)

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@tag"]) {
          query: Query
        }

        type A {
          a: String!
        }

        type B {
          b: String!
        }

        type Query {
          _service: _Service!
          hello: Union!
        }

        union Union @tag(name: "myTag") @tag(name: "anotherTag") = A | B

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()


def test_tag_printed_correctly_on_inputs():
    @strawberry.federation.input(tags=["myTag", "anotherTag"])
    class Input:
        a: str = strawberry.federation.field(tags=["myTag", "anotherTag"])

    @strawberry.federation.type
    class Query:
        hello: str

    schema = strawberry.federation.Schema(query=Query, types=[Input])

    expected = """
        schema @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@tag"]) {
          query: Query
        }

        input Input @tag(name: "myTag") @tag(name: "anotherTag") {
          a: String! @tag(name: "myTag") @tag(name: "anotherTag")
        }

        type Query {
          _service: _Service!
          hello: String!
        }

        scalar _Any

        type _Service {
          sdl: String!
        }
    """

    assert schema.as_str() == textwrap.dedent(expected).strip()



import datetime

import pytest
from graphql import GraphQLError

import strawberry
from strawberry.types.execution import ExecutionResult


def test_serialization():
    @strawberry.type
    class Query:
        @strawberry.field
        def serialize(self) -> datetime.time:
            return datetime.time(13, 37)

    schema = strawberry.Schema(Query)

    result = schema.execute_sync("{ serialize }")

    assert not result.errors
    assert result.data["serialize"] == "13:37:00"


def test_deserialization():
    @strawberry.type
    class Query:
        deserialized = None

        @strawberry.field
        def deserialize(self, arg: datetime.time) -> bool:
            Query.deserialized = arg
            return True

    schema = strawberry.Schema(Query)

    query = """query Deserialize($value: Time!) {
        deserialize(arg: $value)
    }"""
    result = schema.execute_sync(query, variable_values={"value": "13:37:00"})

    assert not result.errors
    assert Query.deserialized == datetime.time(13, 37)


def test_deserialization_with_parse_literal():
    @strawberry.type
    class Query:
        deserialized = None

        @strawberry.field
        def deserialize(self, arg: datetime.time) -> bool:
            Query.deserialized = arg
            return True

    schema = strawberry.Schema(Query)

    query = """query Deserialize {
        deserialize(arg: "13:37:00")
    }"""
    result = schema.execute_sync(query)

    assert not result.errors
    assert Query.deserialized == datetime.time(13, 37)


def execute_mutation(value) -> ExecutionResult:
    @strawberry.type
    class Query:
        ok: bool

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def time_input(self, time_input: datetime.time) -> datetime.time:
            assert isinstance(time_input, datetime.time)
            return time_input

    schema = strawberry.Schema(query=Query, mutation=Mutation)

    return schema.execute_sync(
        """
            mutation timeInput($value: Time!) {
                timeInput(timeInput: $value)
            }
        """,
        variable_values={"value": value},
    )


@pytest.mark.parametrize(
    "value",
    [
        "2012-12-01T09:00",
        "03:30+",
        "03:30+1234567",
        "03:30-25:40",
    ],
)
def test_serialization_of_incorrect_time_string(value):
    """Test GraphQLError is raised for incorrect time.
    The error should exclude "original_error".
    """
    result = execute_mutation(value)
    assert result.errors
    assert isinstance(result.errors[0], GraphQLError)


def test_serialization_error_message_for_incorrect_time_string():
    """Test if error message is using original error message
    from time lib, and is properly formatted.
    """
    result = execute_mutation("25:00")
    assert result.errors
    assert result.errors[0].message.startswith(
        "Variable '$value' got invalid value '25:00'; Value cannot represent a "
        'Time: "25:00". hour must be in 0..23'
    )



import textwrap

import strawberry
from strawberry.schema_directive import Location


def test_schema_directives_and_compose_schema():
    @strawberry.federation.schema_directive(
        locations=[Location.OBJECT],
        name="cacheControl",
        compose=True,
    )
    class CacheControl:
        max_age: int

    @strawberry.federation.schema_directive(
        locations=[Location.OBJECT], name="sensitive"
    )
    class Sensitive:
        reason: str

    @strawberry.federation.type(
        keys=["id"],
        shareable=True,
        extend=True,
        directives=[CacheControl(max_age=42), Sensitive(reason="example")],
    )
    class FederatedType:
        id: strawberry.ID

    @strawberry.type
    class Query:
        federatedType: FederatedType  # noqa: N815

    expected_type = """
    directive @cacheControl(maxAge: Int!) on OBJECT

    directive @sensitive(reason: String!) on OBJECT

    schema @composeDirective(name: "@cacheControl") @link(url: "https://directives.strawberry.rocks/cacheControl/v0.1", import: ["@cacheControl"]) @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@composeDirective", "@key", "@shareable"]) {
      query: Query
    }

    extend type FederatedType @cacheControl(maxAge: 42) @sensitive(reason: "example") @key(fields: "id") @shareable {
      id: ID!
    }

    type Query {
      _entities(representations: [_Any!]!): [_Entity]!
      _service: _Service!
      federatedType: FederatedType!
    }

    scalar _Any

    union _Entity = FederatedType

    type _Service {
      sdl: String!
    }
    """

    schema = strawberry.federation.Schema(
        query=Query,
    )

    assert schema.as_str() == textwrap.dedent(expected_type).strip()


def test_schema_directives_and_compose_schema_custom_import_url():
    @strawberry.federation.schema_directive(
        locations=[Location.OBJECT],
        name="cacheControl",
        compose=True,
        import_url="https://f.strawberry.rocks/cacheControl/v1.0",
    )
    class CacheControl:
        max_age: int

    @strawberry.federation.schema_directive(
        locations=[Location.OBJECT], name="sensitive"
    )
    class Sensitive:
        reason: str

    @strawberry.federation.type(
        keys=["id"],
        shareable=True,
        extend=True,
        directives=[CacheControl(max_age=42), Sensitive(reason="example")],
    )
    class FederatedType:
        id: strawberry.ID

    @strawberry.type
    class Query:
        federatedType: FederatedType  # noqa: N815

    expected_type = """
    directive @cacheControl(maxAge: Int!) on OBJECT

    directive @sensitive(reason: String!) on OBJECT

    schema @composeDirective(name: "@cacheControl") @link(url: "https://f.strawberry.rocks/cacheControl/v1.0", import: ["@cacheControl"]) @link(url: "https://specs.apollo.dev/federation/v2.11", import: ["@composeDirective", "@key", "@shareable"]) {
      query: Query
    }

    extend type FederatedType @cacheControl(maxAge: 42) @sensitive(reason: "example") @key(fields: "id") @shareable {
      id: ID!
    }

    type Query {
      _entities(representations: [_Any!]!): [_Entity]!
      _service: _Service!
      federatedType: FederatedType!
    }

    scalar _Any

    union _Entity = FederatedType

    type _Service {
      sdl: String!
    }
    """

    schema = strawberry.federation.Schema(
        query=Query,
    )

    assert schema.as_str() == textwrap.dedent(expected_type).strip()



import textwrap
from typing import Annotated

import strawberry
from strawberry.field_extensions import InputMutationExtension
from strawberry.schema_directive import Location, schema_directive


@schema_directive(
    locations=[Location.FIELD_DEFINITION],
    name="some_directive",
)
class SomeDirective:
    some: str
    directive: str


@strawberry.type
class Fruit:
    name: str
    color: str


@strawberry.type
class Query:
    @strawberry.field
    def hello(self) -> str:
        return "hi"


@strawberry.type
class Mutation:
    @strawberry.mutation(extensions=[InputMutationExtension()])
    def create_fruit(
        self,
        name: str,
        color: Annotated[
            str,
            strawberry.argument(
                description="The color of the fruit",
                directives=[SomeDirective(some="foo", directive="bar")],
            ),
        ],
        variety: Annotated[
            str | None,
            strawberry.argument(
                deprecation_reason="Use `color` instead",
            ),
        ] = None,
    ) -> Fruit:
        return Fruit(
            name=name,
            color=color,
        )

    @strawberry.mutation(extensions=[InputMutationExtension()])
    async def create_fruit_async(
        self,
        name: str,
        color: Annotated[str, object()],
    ) -> Fruit:
        return Fruit(
            name=name,
            color=color,
        )


schema = strawberry.Schema(query=Query, mutation=Mutation)


def test_schema():
    expected = '''
    directive @some_directive(some: String!, directive: String!) on FIELD_DEFINITION

    input CreateFruitAsyncInput {
      name: String!
      color: String!
    }

    input CreateFruitInput {
      name: String!

      """The color of the fruit"""
      color: String! @some_directive(some: "foo", directive: "bar")
      variety: String = null @deprecated(reason: "Use `color` instead")
    }

    type Fruit {
      name: String!
      color: String!
    }

    type Mutation {
      createFruit(
        """Input data for `createFruit` mutation"""
        input: CreateFruitInput!
      ): Fruit!
      createFruitAsync(
        """Input data for `createFruitAsync` mutation"""
        input: CreateFruitAsyncInput!
      ): Fruit!
    }

    type Query {
      hello: String!
    }
    '''
    assert str(schema).strip() == textwrap.dedent(expected).strip()


def test_input_mutation():
    result = schema.execute_sync(
        """
        mutation TestQuery ($input: CreateFruitInput!) {
            createFruit (input: $input) {
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "input": {
                "name": "Dragonfruit",
                "color": "red",
            }
        },
    )
    assert result.errors is None
    assert result.data == {
        "createFruit": {
            "name": "Dragonfruit",
            "color": "red",
        },
    }


async def test_input_mutation_async():
    result = await schema.execute(
        """
        mutation TestQuery ($input: CreateFruitAsyncInput!) {
            createFruitAsync (input: $input) {
                ... on Fruit {
                    name
                    color
                }
            }
        }
        """,
        variable_values={
            "input": {
                "name": "Dragonfruit",
                "color": "red",
            }
        },
    )
    assert result.errors is None
    assert result.data == {
        "createFruitAsync": {
            "name": "Dragonfruit",
            "color": "red",
        },
    }



from typing import List, Optional

class OperationNameResult:
    optional_int: Optional[int]
    list_of_int: list[int]
    list_of_optional_int: list[Optional[int]]
    optional_list_of_optional_int: Optional[list[Optional[int]]]



from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator
from typing import TYPE_CHECKING

import pytest

from strawberry.subscriptions import GRAPHQL_TRANSPORT_WS_PROTOCOL
from strawberry.subscriptions.protocols.graphql_transport_ws.types import (
    CompleteMessage,
    ConnectionAckMessage,
    ConnectionInitMessage,
    NextMessage,
    SubscribeMessage,
)
from tests.views.schema import schema

if TYPE_CHECKING:
    from channels.testing import WebsocketCommunicator


@pytest.fixture
async def ws() -> AsyncGenerator[WebsocketCommunicator, None]:
    from channels.testing import WebsocketCommunicator

    from strawberry.channels import GraphQLWSConsumer

    client = WebsocketCommunicator(
        GraphQLWSConsumer.as_asgi(schema=schema),
        "/graphql",
        subprotocols=[GRAPHQL_TRANSPORT_WS_PROTOCOL],
    )
    res = await client.connect()
    assert res == (True, GRAPHQL_TRANSPORT_WS_PROTOCOL)

    yield client

    await client.disconnect()


async def test_no_layers():
    from strawberry.channels.handlers.base import ChannelsConsumer

    consumer = ChannelsConsumer()
    # Mimic lack of layers. If layers is not installed/configured in channels,
    # consumer.channel_layer will be `None`
    consumer.channel_layer = None

    msg = (
        "Layers integration is required listening for channels.\n"
        "Check https://channels.readthedocs.io/en/stable/topics/channel_layers.html "
        "for more information"
    )
    with pytest.raises(RuntimeError, match=msg):
        async with consumer.listen_to_channel("foobar"):
            pass


async def test_listen_to_channel_generator_delivers_messages():
    """_listen_to_channel_generator yields messages placed on its queue."""
    from strawberry.channels.handlers.base import ChannelsConsumer

    consumer = ChannelsConsumer()
    queue: asyncio.Queue = asyncio.Queue()
    gen = consumer._listen_to_channel_generator(queue, timeout=None)

    queue.put_nowait({"type": "broadcast", "text": "hello"})
    queue.put_nowait({"type": "broadcast", "text": "world"})

    assert await gen.__anext__() == {"type": "broadcast", "text": "hello"}
    assert await gen.__anext__() == {"type": "broadcast", "text": "world"}

    await gen.aclose()


async def test_listen_to_channel_generator_timeout():
    """_listen_to_channel_generator returns on timeout when no message arrives."""
    from strawberry.channels.handlers.base import ChannelsConsumer

    consumer = ChannelsConsumer()
    queue: asyncio.Queue = asyncio.Queue()
    gen = consumer._listen_to_channel_generator(queue, timeout=0.05)

    # No message on the queue — generator should time out and stop
    with pytest.raises(StopAsyncIteration):
        await gen.__anext__()


async def test_listen_to_channel_generator_aclose_is_clean():
    """Closing the generator after receiving a message must not raise."""
    from strawberry.channels.handlers.base import ChannelsConsumer

    consumer = ChannelsConsumer()
    queue: asyncio.Queue = asyncio.Queue()
    gen = consumer._listen_to_channel_generator(queue, timeout=None)

    queue.put_nowait({"type": "broadcast", "text": "msg"})
    await gen.__anext__()

    # Generator is now suspended after yielding. Close it cleanly.
    await gen.aclose()


async def test_listen_to_channel_generator_throw_timeout_at_yield_propagates():
    """TimeoutError thrown while the generator is suspended at its yield
    must propagate, not be swallowed by the internal timeout handler.

    Before the fix, ``yield await awaitable`` was a compound expression
    whose ``yield`` fell inside the ``try/except asyncio.TimeoutError``
    block (per the bytecode exception table). A ``TimeoutError`` thrown
    at the yield point was therefore incorrectly caught, silently
    stopping the generator instead of letting the caller handle it.
    """
    from strawberry.channels.handlers.base import ChannelsConsumer

    consumer = ChannelsConsumer()
    queue: asyncio.Queue = asyncio.Queue()
    gen = consumer._listen_to_channel_generator(queue, timeout=None)

    queue.put_nowait({"type": "broadcast", "text": "msg"})
    await gen.__anext__()  # generator is now suspended at its yield

    # Throw TimeoutError — it should propagate, not be caught internally
    with pytest.raises(asyncio.TimeoutError):
        await gen.athrow(asyncio.TimeoutError())


@pytest.mark.django_db
async def test_listen_to_channel_timeout(ws: WebsocketCommunicator):
    from channels.layers import get_channel_layer

    await ws.send_json_to(ConnectionInitMessage({"type": "connection_init"}))

    connection_ack_message: ConnectionAckMessage = await ws.receive_json_from()
    assert connection_ack_message == {"type": "connection_ack"}

    await ws.send_json_to(
        SubscribeMessage(
            {
                "id": "sub1",
                "type": "subscribe",
                "payload": {
                    "query": "subscription { listenerWithConfirmation(timeout: 0.5) }",
                },
            }
        )
    )

    channel_layer = get_channel_layer()
    assert channel_layer

    next_message1: NextMessage = await ws.receive_json_from()
    assert "data" in next_message1["payload"]
    assert next_message1["payload"]["data"] is not None
    confirmation = next_message1["payload"]["data"]["listenerWithConfirmation"]
    assert confirmation is None

    next_message2 = await ws.receive_json_from()
    assert "data" in next_message2["payload"]
    assert next_message2["payload"]["data"] is not None
    channel_name = next_message2["payload"]["data"]["listenerWithConfirmation"]
    assert channel_name

    complete_message: CompleteMessage = await ws.receive_json_from()
    assert complete_message == {"id": "sub1", "type": "complete"}


@pytest.mark.django_db
async def test_listen_to_channel_no_message_on_channel(ws: WebsocketCommunicator):
    from channels.layers import get_channel_layer

    await ws.send_json_to(ConnectionInitMessage({"type": "connection_init"}))

    connection_ack_message: ConnectionAckMessage = await ws.receive_json_from()
    assert connection_ack_message == {"type": "connection_ack"}

    await ws.send_json_to(
        SubscribeMessage(
            {
                "id": "sub1",
                "type": "subscribe",
                "payload": {
                    "query": "subscription { listenerWithConfirmation(timeout: 0.5) }",
                },
            }
        )
    )

    channel_layer = get_channel_layer()
    assert channel_layer

    next_message1: NextMessage = await ws.receive_json_from()
    assert "data" in next_message1["payload"]
    assert next_message1["payload"]["data"] is not None
    confirmation = next_message1["payload"]["data"]["listenerWithConfirmation"]
    assert confirmation is None

    next_message2 = await ws.receive_json_from()
    assert "data" in next_message2["payload"]
    assert next_message2["payload"]["data"] is not None
    channel_name = next_message2["payload"]["data"]["listenerWithConfirmation"]
    assert channel_name

    await channel_layer.send(
        "totally-not-out-channel",
        {
            "type": "test.message",
            "text": "Hello there!",
        },
    )

    complete_message: CompleteMessage = await ws.receive_json_from()
    assert complete_message == {"id": "sub1", "type": "complete"}


@pytest.mark.django_db
async def test_listen_to_channel_group(ws: WebsocketCommunicator):
    from channels.layers import get_channel_layer

    await ws.send_json_to(ConnectionInitMessage({"type": "connection_init"}))

    connection_ack_message: ConnectionAckMessage = await ws.receive_json_from()
    assert connection_ack_message == {"type": "connection_ack"}

    await ws.send_json_to(
        SubscribeMessage(
            {
                "id": "sub1",
                "type": "subscribe",
                "payload": {
                    "query": 'subscription { listenerWithConfirmation(group: "foobar") }',
                },
            }
        )
    )

    channel_layer = get_channel_layer()
    assert channel_layer

    next_message1: NextMessage = await ws.receive_json_from()
    assert "data" in next_message1["payload"]
    assert next_message1["payload"]["data"] is not None
    confirmation = next_message1["payload"]["data"]["listenerWithConfirmation"]
    assert confirmation is None

    next_message2 = await ws.receive_json_from()
    assert "data" in next_message2["payload"]
    assert next_message2["payload"]["data"] is not None
    channel_name = next_message2["payload"]["data"]["listenerWithConfirmation"]

    # Sent at least once to the consumer to make sure the groups were registered
    await channel_layer.send(
        channel_name,
        {
            "type": "test.message",
            "text": "Hello there!",
        },
    )

    next_message3: NextMessage = await ws.receive_json_from()
    assert next_message3 == {
        "id": "sub1",
        "type": "next",
        "payload": {
            "data": {"listenerWithConfirmation": "Hello there!"},
            "extensions": {"example": "example"},
        },
    }

    await channel_layer.group_send(
        "foobar",
        {
            "type": "test.message",
            "text": "Hello there!",
        },
    )

    next_message4: NextMessage = await ws.receive_json_from()
    assert next_message4 == {
        "id": "sub1",
        "type": "next",
        "payload": {
            "data": {"listenerWithConfirmation": "Hello there!"},
            "extensions": {"example": "example"},
        },
    }

    await ws.send_json_to(CompleteMessage({"id": "sub1", "type": "complete"}))


@pytest.mark.django_db
async def test_listen_to_channel_group_twice(ws: WebsocketCommunicator):
    from channels.layers import get_channel_layer

    await ws.send_json_to(ConnectionInitMessage({"type": "connection_init"}))

    connection_ack_message: ConnectionAckMessage = await ws.receive_json_from()
    assert connection_ack_message == {"type": "connection_ack"}

    await ws.send_json_to(
        SubscribeMessage(
            {
                "id": "sub1",
                "type": "subscribe",
                "payload": {
                    "query": 'subscription { listenerWithConfirmation(group: "group1") }',
                },
            }
        )
    )

    await ws.send_json_to(
        SubscribeMessage(
            {
                "id": "sub2",
                "type": "subscribe",
                "payload": {
                    "query": 'subscription { listenerWithConfirmation(group: "group2") }',
                },
            }
        )
    )

    channel_layer = get_channel_layer()
    assert channel_layer

    # Wait for confirmation for channel subscriptions
    messages = await asyncio.gather(
        ws.receive_json_from(),
        ws.receive_json_from(),
        ws.receive_json_from(),
        ws.receive_json_from(),
    )
    confirmation1 = next(
        i
        for i in messages
        if not i["payload"]["data"]["listenerWithConfirmation"] and i["id"] == "sub1"
    )
    confirmation2 = next(
        i
        for i in messages
        if not i["payload"]["data"]["listenerWithConfirmation"] and i["id"] == "sub2"
    )
    channel_name1 = next(
        i
        for i in messages
        if i["payload"]["data"]["listenerWithConfirmation"] and i["id"] == "sub1"
    )
    channel_name2 = next(
        i
        for i in messages
        if i["payload"]["data"]["listenerWithConfirmation"] and i["id"] == "sub2"
    )

    # Ensure correct ordering of responses
    assert messages.index(confirmation1) < messages.index(channel_name1)
    assert messages.index(confirmation2) < messages.index(channel_name2)
    channel_name = channel_name1["payload"]["data"]["listenerWithConfirmation"]

    # Sent at least once to the consumer to make sure the groups were registered
    await channel_layer.send(
        channel_name,
        {
            "type": "test.message",
            "text": "Hello there!",
        },
    )

    next_message1: NextMessage = await ws.receive_json_from()
    next_message2: NextMessage = await ws.receive_json_from()
    assert {"sub1", "sub2"} == {next_message1["id"], next_message2["id"]}

    assert "data" in next_message1["payload"]
    assert next_message1["payload"]["data"] is not None
    assert (
        next_message1["payload"]["data"]["listenerWithConfirmation"] == "Hello there!"
    )

    assert "data" in next_message2["payload"]
    assert next_message2["payload"]["data"] is not None
    assert (
        next_message2["payload"]["data"]["listenerWithConfirmation"] == "Hello there!"
    )

    # We now have two listen_to_channel AsyncGenerators waiting, one for id="sub1"
    # and one for id="sub2". This group message will be received by both of them
    # as they are both running on the same ChannelsConsumer instance so even
    # though "sub2" was initialised with "group2" as the argument, it will receive
    # this message for "group1"
    await channel_layer.group_send(
        "group1",
        {
            "type": "test.message",
            "text": "Hello group 1!",
        },
    )

    next_message3: NextMessage = await ws.receive_json_from()
    next_message4: NextMessage = await ws.receive_json_from()
    assert {"sub1", "sub2"} == {next_message3["id"], next_message4["id"]}

    assert "data" in next_message3["payload"]
    assert next_message3["payload"]["data"] is not None
    assert (
        next_message3["payload"]["data"]["listenerWithConfirmation"] == "Hello group 1!"
    )

    assert "data" in next_message4["payload"]
    assert next_message4["payload"]["data"] is not None
    assert (
        next_message4["payload"]["data"]["listenerWithConfirmation"] == "Hello group 1!"
    )

    await channel_layer.group_send(
        "group2",
        {
            "type": "test.message",
            "text": "Hello group 2!",
        },
    )

    next_message5: NextMessage = await ws.receive_json_from()
    next_message6: NextMessage = await ws.receive_json_from()
    assert {"sub1", "sub2"} == {next_message5["id"], next_message6["id"]}

    assert "data" in next_message5["payload"]
    assert next_message5["payload"]["data"] is not None
    assert (
        next_message5["payload"]["data"]["listenerWithConfirmation"] == "Hello group 2!"
    )

    assert "data" in next_message6["payload"]
    assert next_message6["payload"]["data"] is not None
    assert (
        next_message6["payload"]["data"]["listenerWithConfirmation"] == "Hello group 2!"
    )

    await ws.send_json_to(CompleteMessage({"id": "sub1", "type": "complete"}))
    await ws.send_json_to(CompleteMessage({"id": "sub2", "type": "complete"}))



from inline_snapshot import snapshot

from .utils.marks import requires_mypy, requires_pyright, requires_ty, skip_on_windows
from .utils.typecheck import Result, typecheck

pytestmark = [skip_on_windows, requires_pyright, requires_mypy, requires_ty]


CODE = """
import strawberry


@strawberry.type
class User:
    name: str
    age: strawberry.Private[int]


patrick = User(name="Patrick", age=1)
User(n="Patrick")

reveal_type(patrick.name)
reveal_type(patrick.age)
"""


def test():
    results = typecheck(CODE)

    assert results.pyright == snapshot(
        [
            Result(
                type="error",
                message='Arguments missing for parameters "name", "age"',
                line=12,
                column=1,
            ),
            Result(type="error", message='No parameter named "n"', line=12, column=6),
            Result(
                type="information",
                message='Type of "patrick.name" is "str"',
                line=14,
                column=13,
            ),
            Result(
                type="information",
                message='Type of "patrick.age" is "int"',
                line=15,
                column=13,
            ),
        ]
    )
    assert results.mypy == snapshot(
        [
            Result(
                type="error",
                message='Unexpected keyword argument "n" for "User"',
                line=12,
                column=1,
            ),
            Result(
                type="note",
                message='Revealed type is "str"',
                line=14,
                column=13,
            ),
            Result(
                type="note",
                message='Revealed type is "int"',
                line=15,
                column=13,
            ),
        ]
    )
    assert results.ty == snapshot(
        [
            Result(
                type="error",
                message="No arguments provided for required parameters `name`, `age`",
                line=12,
                column=1,
            ),
            Result(
                type="error",
                message="Argument `n` does not match any known parameter",
                line=12,
                column=6,
            ),
            Result(
                type="information",
                message="Revealed type: `str`",
                line=14,
                column=13,
            ),
            Result(
                type="information",
                message="Revealed type: `int`",
                line=15,
                column=13,
            ),
        ]
    )



from __future__ import annotations

from collections.abc import AsyncGenerator
from typing import TYPE_CHECKING, Any

import pytest

from strawberry.subscriptions import GRAPHQL_TRANSPORT_WS_PROTOCOL, GRAPHQL_WS_PROTOCOL
from tests.views.schema import schema

if TYPE_CHECKING:
    from strawberry.channels.testing import GraphQLWebsocketCommunicator


@pytest.fixture(params=[GRAPHQL_TRANSPORT_WS_PROTOCOL, GRAPHQL_WS_PROTOCOL])
async def communicator(
    request: Any,
) -> AsyncGenerator[GraphQLWebsocketCommunicator, None]:
    from strawberry.channels import GraphQLWSConsumer
    from strawberry.channels.testing import GraphQLWebsocketCommunicator

    application = GraphQLWSConsumer.as_asgi(schema=schema, keep_alive_interval=50)

    async with GraphQLWebsocketCommunicator(
        protocol=request.param,
        application=application,
        path="/graphql",
        connection_params={"strawberry": "Hi"},
    ) as client:
        yield client


async def test_simple_subscribe(communicator: GraphQLWebsocketCommunicator):
    async for res in communicator.subscribe(
        query='subscription { echo(message: "Hi") }'
    ):
        assert res.data == {"echo": "Hi"}


async def test_subscribe_unexpected_error(communicator):
    async for res in communicator.subscribe(
        query='subscription { exception(message: "Hi") }'
    ):
        assert res.errors[0].message == "Hi"


async def test_graphql_error(communicator):
    async for res in communicator.subscribe(
        query='subscription { error(message: "Hi") }'
    ):
        assert res.errors[0].message == "Hi"


async def test_simple_connection_params(communicator):
    async for res in communicator.subscribe(query="subscription { connectionParams }"):
        assert res.data["connectionParams"]["strawberry"] == "Hi"



from unittest.mock import Mock

import pytest
from graphql.error import GraphQLError

import strawberry
from strawberry.extensions import MaskErrors


def test_mask_all_errors():
    @strawberry.type
    class Query:
        @strawberry.field
        def hidden_error(self) -> str:
            raise KeyError("This error is not visible")

    schema = strawberry.Schema(query=Query, extensions=[MaskErrors()])

    query = "query { hiddenError }"

    result = schema.execute_sync(query)
    assert result.errors is not None
    formatted_errors = [err.formatted for err in result.errors]
    assert formatted_errors == [
        {
            "locations": [{"column": 9, "line": 1}],
            "message": "Unexpected error.",
            "path": ["hiddenError"],
        }
    ]


@pytest.mark.asyncio
async def test_mask_all_errors_async():
    @strawberry.type
    class Query:
        @strawberry.field
        def hidden_error(self) -> str:
            raise KeyError("This error is not visible")

    schema = strawberry.Schema(query=Query, extensions=[MaskErrors()])

    query = "query { hiddenError }"

    result = await schema.execute(query)
    assert result.errors is not None
    formatted_errors = [err.formatted for err in result.errors]
    assert formatted_errors == [
        {
            "locations": [{"column": 9, "line": 1}],
            "message": "Unexpected error.",
            "path": ["hiddenError"],
        }
    ]


def test_mask_some_errors():
    class VisibleError(Exception):
        pass

    @strawberry.type
    class Query:
        @strawberry.field
        def visible_error(self) -> str:
            raise VisibleError("This error is visible")

        @strawberry.field
        def hidden_error(self) -> str:
            raise Exception("This error is not visible")

    def should_mask_error(error: GraphQLError) -> bool:
        original_error = error.original_error
        return not (original_error and isinstance(original_error, VisibleError))

    schema = strawberry.Schema(
        query=Query, extensions=[MaskErrors(should_mask_error=should_mask_error)]
    )

    query = "query { hiddenError }"

    result = schema.execute_sync(query)
    assert result.errors is not None
    formatted_errors = [err.formatted for err in result.errors]
    assert formatted_errors == [
        {
            "locations": [{"column": 9, "line": 1}],
            "message": "Unexpected error.",
            "path": ["hiddenError"],
        }
    ]

    query = "query { visibleError }"

    result = schema.execute_sync(query)
    assert result.errors is not None
    formatted_errors = [err.formatted for err in result.errors]
    assert formatted_errors == [
        {
            "locations": [{"column": 9, "line": 1}],
            "message": "This error is visible",
            "path": ["visibleError"],
        }
    ]


def test_process_errors_original_error():
    @strawberry.type
    class Query:
        @strawberry.field
        def hidden_error(self) -> str:
            raise ValueError("This error is not visible")

    mock_process_error = Mock()

    class CustomSchema(strawberry.Schema):
        def process_errors(self, errors, execution_context):
            for error in errors:
                mock_process_error(error)

    schema = CustomSchema(query=Query, extensions=[MaskErrors()])

    query = "query { hiddenError }"

    result = schema.execute_sync(query)
    assert result.errors is not None
    formatted_errors = [err.formatted for err in result.errors]
    assert formatted_errors == [
        {
            "locations": [{"column": 9, "line": 1}],
            "message": "Unexpected error.",
            "path": ["hiddenError"],
        }
    ]

    assert mock_process_error.call_count == 1
    call = mock_process_error.call_args_list[0]
    assert call[0][0].message == "This error is not visible"
    assert isinstance(call[0][0].original_error, ValueError)


def test_graphql_error_masking():
    @strawberry.type
    class Query:
        @strawberry.field
        def graphql_error(self) -> str:
            return None  # type: ignore

    schema = strawberry.Schema(query=Query, extensions=[MaskErrors()])

    query = "query { graphqlError }"

    result = schema.execute_sync(query)
    assert result.errors is not None
    formatted_errors = [err.formatted for err in result.errors]
    assert formatted_errors == [
        {
            "locations": [{"column": 9, "line": 1}],
            "message": "Unexpected error.",
            "path": ["graphqlError"],
        }
    ]


@pytest.mark.asyncio
async def test_mask_errors_with_strawberry_execution_result_async():
    """Test that MaskErrors works correctly with async execution (which returns StrawberryExecutionResult)."""

    @strawberry.type
    class Query:
        @strawberry.field
        def test_field(self) -> str:
            raise ValueError("Original error message")

    schema = strawberry.Schema(query=Query, extensions=[MaskErrors()])

    query = "query { testField }"

    # Use async execution to ensure we get StrawberryExecutionResult path
    result = await schema.execute(query)
    assert result.errors is not None
    formatted_errors = [err.formatted for err in result.errors]
    assert formatted_errors == [
        {
            "locations": [{"column": 9, "line": 1}],
            "message": "Unexpected error.",
            "path": ["testField"],
        }
    ]


@pytest.mark.asyncio
async def test_mask_errors_selective_async():
    """Test selective error masking with async execution."""

    class VisibleError(Exception):
        pass

    @strawberry.type
    class Query:
        @strawberry.field
        def visible_error(self) -> str:
            raise VisibleError("This error is visible")

        @strawberry.field
        def hidden_error(self) -> str:
            raise ValueError("This error is not visible")

    def should_mask_error(error: GraphQLError) -> bool:
        original_error = error.original_error
        return not (original_error and isinstance(original_error, VisibleError))

    schema = strawberry.Schema(
        query=Query, extensions=[MaskErrors(should_mask_error=should_mask_error)]
    )

    # Test hidden error
    query = "query { hiddenError }"
    result = await schema.execute(query)
    assert result.errors is not None
    formatted_errors = [err.formatted for err in result.errors]
    assert formatted_errors == [
        {
            "locations": [{"column": 9, "line": 1}],
            "message": "Unexpected error.",
            "path": ["hiddenError"],
        }
    ]

    # Test visible error
    query = "query { visibleError }"
    result = await schema.execute(query)
    assert result.errors is not None
    formatted_errors = [err.formatted for err in result.errors]
    assert formatted_errors == [
        {
            "locations": [{"column": 9, "line": 1}],
            "message": "This error is visible",
            "path": ["visibleError"],
        }
    ]



from typing import TYPE_CHECKING, Annotated

import strawberry

if TYPE_CHECKING:
    from .type_b import TypeB


@strawberry.type
class TypeA:
    list_of_b: (
        list[Annotated["TypeB", strawberry.lazy("tests.schema.test_lazy_types.type_b")]]
        | None
    ) = None

    @strawberry.field
    def type_b(self) -> Annotated["TypeB", strawberry.lazy(".type_b")]:
        from .type_b import TypeB

        return TypeB()



import warnings
from enum import Enum
from typing import Optional, TypeVar

import pytest

import strawberry
from strawberry.types.info import Info


def test_enum():
    @strawberry.enum
    class Locale(Enum):
        UNITED_STATES = "en_US"
        UK = "en_UK"
        AUSTRALIA = "en_AU"

    @strawberry.mutation
    def set_locale(locale: Locale) -> bool:
        _ = locale
        return True

    argument = set_locale.arguments[0]
    # TODO: Remove reference to .__strawberry_definition__ with StrawberryEnumDefinition
    assert argument.type is Locale.__strawberry_definition__


def test_forward_reference():
    global SearchInput

    @strawberry.field
    def search(search_input: "SearchInput") -> bool:
        _ = search_input
        return True

    @strawberry.input
    class SearchInput:
        query: str

    argument = search.arguments[0]
    assert argument.type is SearchInput

    del SearchInput


def test_list():
    @strawberry.field
    def get_longest_word(words: list[str]) -> str:
        _ = words
        return "I cheated"

    argument = get_longest_word.arguments[0]
    assert argument.type == list[str]


def test_literal():
    @strawberry.field
    def get_name(id_: int) -> str:
        _ = id_
        return "Lord Buckethead"

    argument = get_name.arguments[0]
    assert argument.type is int


def test_object():
    @strawberry.type
    class PersonInput:
        proper_noun: bool

    @strawberry.field
    def get_id(person_input: PersonInput) -> int:
        _ = person_input
        return 0

    argument = get_id.arguments[0]
    assert argument.type is PersonInput


def test_optional():
    @strawberry.field
    def set_age(age: int | None) -> bool:
        _ = age
        return True

    argument = set_age.arguments[0]
    assert argument.type == Optional[int]


def test_type_var():
    T = TypeVar("T")

    @strawberry.field
    def set_value(value: T) -> bool:
        _ = value
        return True

    argument = set_value.arguments[0]
    assert argument.type == T


ContextType = TypeVar("ContextType")
RootValueType = TypeVar("RootValueType")


class CustomInfo(Info[ContextType, RootValueType]):
    """Subclassed Info type used to test dependency injection."""


@pytest.mark.parametrize(
    "annotation",
    [CustomInfo, CustomInfo[None, None], Info, Info[None, None]],
)
def test_custom_info(annotation):
    """Test to ensure that subclassed Info does not raise warning."""
    with warnings.catch_warnings():
        warnings.filterwarnings("error")

        def get_info(info) -> bool:
            _ = info
            return True

        get_info.__annotations__["info"] = annotation
        get_info_field = strawberry.field(get_info)

        assert not get_info_field.arguments  # Should have no arguments matched

        info_parameter = get_info_field.base_resolver.info_parameter
        assert info_parameter is not None
        assert info_parameter.name == "info"



from pathlib import Path

from pytest_snapshot.plugin import Snapshot
from typer import Typer
from typer.testing import CliRunner

HERE = Path(__file__).parent


def test_upgrade_returns_error_code_if_codemod_does_not_exist(
    cli_app: Typer, cli_runner: CliRunner
):
    result = cli_runner.invoke(
        cli_app,
        ["upgrade", "a_random_codemod", "."],
    )

    assert result.exit_code == 2
    assert 'Upgrade named "a_random_codemod" does not exist' in result.stdout


def test_upgrade_works_annotated_unions(
    cli_app: Typer, cli_runner: CliRunner, tmp_path: Path, snapshot: Snapshot
):
    source = HERE / "fixtures/unions.py"

    target = tmp_path / "unions.py"
    target.write_text(source.read_text())

    result = cli_runner.invoke(
        cli_app,
        ["upgrade", "--python-target", "3.11", "annotated-union", str(target)],
    )

    assert result.exit_code == 1
    assert "1 files changed\n  - 0 files skipped" in result.stdout

    snapshot.snapshot_dir = HERE / "snapshots"
    snapshot.assert_match(target.read_text(), "unions.py")


def test_upgrade_works_annotated_unions_typing_extensions(
    cli_app: Typer, cli_runner: CliRunner, tmp_path: Path, snapshot: Snapshot
):
    source = HERE / "fixtures/unions.py"

    target = tmp_path / "unions.py"
    target.write_text(source.read_text())

    result = cli_runner.invoke(
        cli_app,
        [
            "upgrade",
            "--use-typing-extensions",
            "--python-target",
            "3.11",
            "annotated-union",
            str(target),
        ],
    )

    assert result.exit_code == 1
    assert "1 files changed\n  - 0 files skipped" in result.stdout

    snapshot.snapshot_dir = HERE / "snapshots"
    snapshot.assert_match(target.read_text(), "unions_typing_extension.py")



import strawberry
from typing import Annotated

# create a few types and then a union type


@strawberry.type
class Foo:
    a: str


@strawberry.type
class Bar:
    b: str


@strawberry.type
class Baz:
    c: str


@strawberry.type
class Qux:
    d: str


# this is the union type

Union1 = Annotated[Foo | Bar | Baz | Qux, strawberry.union(name="Union1")]
Union2 = Annotated[Baz | Qux, strawberry.union(name="Union2")]



import textwrap
from typing import Annotated, Union

import strawberry
from strawberry.printer import print_schema


@strawberry.type
class TypeA:
    a: int


@strawberry.type
class TypeB:
    b: int


ABUnion = Annotated[TypeA | TypeB, strawberry.union("ABUnion")]

# New syntax unions (without explicit types= argument) - for issue #4066
NewSyntaxUnion = Annotated[TypeA | TypeB, strawberry.union("NewSyntaxUnion")]
NewSyntaxUnionTyping = Annotated[
    Union[TypeA, TypeB], strawberry.union("NewSyntaxUnionTyping")
]


TypeALazy = Annotated[
    "TypeA", strawberry.lazy("tests.schema.test_lazy_types.test_lazy_unions")
]
TypeBLazy = Annotated[
    "TypeB", strawberry.lazy("tests.schema.test_lazy_types.test_lazy_unions")
]
LazyABUnion = Annotated[
    TypeALazy | TypeBLazy,
    strawberry.union("LazyABUnion"),
]


def test_lazy_union_with_non_lazy_members():
    @strawberry.type
    class Query:
        ab: Annotated[
            "ABUnion", strawberry.lazy("tests.schema.test_lazy_types.test_lazy_unions")
        ]

    expected = """
        union ABUnion = TypeA | TypeB

        type Query {
          ab: ABUnion!
        }

        type TypeA {
          a: Int!
        }

        type TypeB {
          b: Int!
        }
    """

    schema = strawberry.Schema(query=Query)
    assert print_schema(schema) == textwrap.dedent(expected).strip()


def test_lazy_union_with_lazy_members():
    @strawberry.type
    class Query:
        ab: Annotated[
            "LazyABUnion",
            strawberry.lazy("tests.schema.test_lazy_types.test_lazy_unions"),
        ]

    expected = """
        union LazyABUnion = TypeA | TypeB

        type Query {
          ab: LazyABUnion!
        }

        type TypeA {
          a: Int!
        }

        type TypeB {
          b: Int!
        }
    """

    schema = strawberry.Schema(query=Query)
    assert print_schema(schema) == textwrap.dedent(expected).strip()


def test_lazy_union_new_syntax_pipe():
    """Test lazy reference to a union defined with the new pipe syntax.

    This tests the fix for issue #4066 where lazy unions using the new syntax
    (without types= argument) would fail with "must define one or more member types".
    """

    @strawberry.type
    class Query:
        item: Annotated[
            "NewSyntaxUnion",
            strawberry.lazy("tests.schema.test_lazy_types.test_lazy_unions"),
        ]

    expected = """
        union NewSyntaxUnion = TypeA | TypeB

        type Query {
          item: NewSyntaxUnion!
        }

        type TypeA {
          a: Int!
        }

        type TypeB {
          b: Int!
        }
    """

    schema = strawberry.Schema(query=Query)
    assert print_schema(schema) == textwrap.dedent(expected).strip()


def test_lazy_union_new_syntax_typing_union():
    """Test lazy reference to a union defined with typing.Union syntax."""

    @strawberry.type
    class Query:
        item: Annotated[
            "NewSyntaxUnionTyping",
            strawberry.lazy("tests.schema.test_lazy_types.test_lazy_unions"),
        ]

    expected = """
        union NewSyntaxUnionTyping = TypeA | TypeB

        type Query {
          item: NewSyntaxUnionTyping!
        }

        type TypeA {
          a: Int!
        }

        type TypeB {
          b: Int!
        }
    """

    schema = strawberry.Schema(query=Query)
    assert print_schema(schema) == textwrap.dedent(expected).strip()



import textwrap

import pydantic
import pytest
from pydantic.version import VERSION as PYDANTIC_VERSION

import strawberry

IS_PYDANTIC_V2: bool = PYDANTIC_VERSION.startswith("2.")

if IS_PYDANTIC_V2:
    from pydantic import computed_field


@pytest.mark.skipif(
    not IS_PYDANTIC_V2, reason="Requires Pydantic v2 for computed_field"
)
def test_computed_field():
    class UserModel(pydantic.BaseModel):
        age: int

        @computed_field
        @property
        def next_age(self) -> int:
            return self.age + 1

    @strawberry.experimental.pydantic.type(
        UserModel, all_fields=True, include_computed=True
    )
    class User:
        pass

    @strawberry.experimental.pydantic.type(UserModel, all_fields=True)
    class UserNoComputed:
        pass

    @strawberry.type
    class Query:
        @strawberry.field
        def user(self) -> User:
            return User.from_pydantic(UserModel(age=1))

        @strawberry.field
        def user_no_computed(self) -> UserNoComputed:
            return UserNoComputed.from_pydantic(UserModel(age=1))

    schema = strawberry.Schema(query=Query)

    expected_schema = """
    type Query {
      user: User!
      userNoComputed: UserNoComputed!
    }

    type User {
      age: Int!
      nextAge: Int!
    }

    type UserNoComputed {
      age: Int!
    }
    """

    assert str(schema) == textwrap.dedent(expected_schema).strip()

    query = "{ user { age nextAge } }"

    result = schema.execute_sync(query)
    assert not result.errors
    assert result.data["user"]["age"] == 1
    assert result.data["user"]["nextAge"] == 2



from enum import Enum
from typing import Optional, TypeVar, Union

from asgiref.sync import sync_to_async

import strawberry
from strawberry.types.fields.resolver import StrawberryResolver


def test_enum():
    @strawberry.enum
    class Language(Enum):
        ENGLISH = "english"
        ITALIAN = "italian"
        JAPANESE = "japanese"

    def get_spoken_language() -> Language:
        return Language.ENGLISH

    resolver = StrawberryResolver(get_spoken_language)
    # TODO: Remove reference to .__strawberry_definition__ with StrawberryEnumDefinition
    assert resolver.type is Language.__strawberry_definition__


def test_forward_references():
    global FutureUmpire

    def get_sportsball_official() -> "FutureUmpire":
        return FutureUmpire("ref")  # noqa: F821

    @strawberry.type
    class FutureUmpire:
        name: str

    resolver = StrawberryResolver(get_sportsball_official)
    assert resolver.type is FutureUmpire

    del FutureUmpire


def test_list():
    def get_collection_types() -> list[str]:
        return ["list", "tuple", "dict", "set"]

    resolver = StrawberryResolver(get_collection_types)
    assert resolver.type == list[str]


def test_literal():
    def version() -> float:
        return 1.0

    resolver = StrawberryResolver(version)
    assert resolver.type is float


def test_object():
    @strawberry.type
    class Polygon:
        edges: int
        faces: int

    def get_2d_object() -> Polygon:
        return Polygon(12, 6)

    resolver = StrawberryResolver(get_2d_object)
    assert resolver.type is Polygon


def test_optional():
    def stock_market_tool() -> str | None: ...

    resolver = StrawberryResolver(stock_market_tool)
    assert resolver.type == Optional[str]


def test_type_var():
    T = TypeVar("T")

    def caffeinated_drink() -> T: ...

    resolver = StrawberryResolver(caffeinated_drink)
    assert resolver.type == T


def test_union():
    @strawberry.type
    class Venn:
        foo: int

    @strawberry.type
    class Diagram:
        bar: float

    def get_overlap() -> Venn | Diagram: ...

    resolver = StrawberryResolver(get_overlap)
    assert resolver.type == Union[Venn, Diagram]


def test_sync_to_async_resolver():
    @sync_to_async
    def async_resolver() -> str:
        return "patrick"

    resolver = StrawberryResolver(async_resolver)
    assert resolver.is_async
